from typing import Any, Optional, Mapping, Dict, Iterator

from mincepy import archives
from mincepy import historians  # pylint: disable=unused-import
from mincepy import exceptions

__all__ = ('Meta',)


class Meta:
    """A class for grouping metadata related methods"""

    # Meta is a 'friend' of Historian and so can access privates pylint: disable=protected-access

    def __init__(self, historian, archive):
        self._hist = historian  # type: historians.Historian
        self._archive = archive
        self._sticky = {}

    @property
    def sticky(self) -> dict:
        return self._sticky

    def get(self, obj_or_identifier) -> Optional[dict]:
        """Get the metadata for an object

        :param obj_or_identifier: either the object instance, an object ID or a snapshot reference
        """
        results = self.get_many((obj_or_identifier,))
        assert len(results) == 1
        meta = tuple(results.values())[0]
        return meta

    def get_many(self, obj_or_identifiers) -> Dict[Any, dict]:
        obj_ids = set(map(self._hist._ensure_obj_id, obj_or_identifiers))
        trans = self._hist.current_transaction()
        if trans:
            # First, get what we can from the archive
            found = {}
            for obj_id in obj_ids:
                try:
                    found[obj_id] = trans.get_meta(obj_id)
                except exceptions.NotFound:
                    pass

            # Now get anything else from the archive
            obj_ids -= found.keys()
            if obj_ids:
                from_archive = self._archive.meta_get_many(obj_ids)
                # Now put into the transaction so it doesn't look it up again.
                for obj_id in obj_ids:
                    trans.set_meta(obj_id, from_archive[obj_id])
                found.update(from_archive)

            return found

        # No transaction
        return self._archive.meta_get_many(obj_ids)

    def set(self, obj_or_identifier, meta: Optional[Mapping]):
        """Set the metadata for an object

        :param obj_or_identifier: either the object instance, an object ID or a snapshot reference
        :param meta: the metadata dictionary
        """
        obj_id = self._hist._ensure_obj_id(obj_or_identifier)
        trans = self._hist.current_transaction()
        if trans:
            return trans.set_meta(obj_id, meta)

        return self._archive.meta_set(obj_id, meta)

    def set_many(self, metas: Mapping[Any, Optional[dict]]):
        mapped = {self._hist._ensure_obj_id(ident): meta for ident, meta in metas.items()}
        trans = self._hist.current_transaction()
        if trans:
            for entry in mapped.items():
                trans.set_meta(*entry)
        else:
            self._archive.meta_set_many(mapped)

    def update(self, obj_or_identifier, meta: Mapping):
        """Update the metadata for an object

        :param obj_or_identifier: either the object instance, an object ID or a snapshot reference
        :param meta: the metadata dictionary
        """
        obj_id = self._hist._ensure_obj_id(obj_or_identifier)
        trans = self._hist.current_transaction()
        if trans:
            # Update the metadata in the transaction
            try:
                current = trans.get_meta(obj_id)
            except exceptions.NotFound:
                current = self._archive.meta_get(obj_id)  # Try the archive
                if current is None:
                    current = {}  # Ok, no meta

            current.update(meta)
            trans.set_meta(obj_id, current)
        else:
            self._archive.meta_update(obj_id, meta)

    def update_many(self, metas: Mapping[Any, Optional[dict]]):
        mapped = {self._hist._ensure_obj_id(ident): meta for ident, meta in metas.items()}
        trans = self._hist.current_transaction()
        if trans:
            for entry in mapped.items():
                self.update(*entry)
        else:
            self._archive.meta_update_many(mapped)

    def find(self, filter, obj_id=None) -> Iterator[archives.Archive.MetaEntry]:  # pylint: disable=redefined-builtin
        """Find metadata matching the given criteria.  Each returned result is a tuple containing
        the corresponding object id and the metadata dictionary itself"""
        return self._archive.meta_find(filter=filter, obj_id=obj_id)

    def distinct(
            self,
            key: str,
            filter: dict = None,  # pylint: disable=redefined-builtin
            obj_id=None) -> Iterator:
        """Yield distinct values found for 'key' within metadata documents, optionally marching a
        search filter.

        The search can optionally be restricted to a set of passed object ids.

        :param key: the document key to get distinct values for
        :param filter: a query filter for the search
        :param obj_id: an optional restriction on the object ids to search.  This ben be either:
            1. a single object id
            2. an iterable of object ids in which is treated as {'$in': list(obj_ids)}
            3. a general query filter to be applied to the object ids
        """
        return self._archive.meta_distinct(key, filter=filter, obj_id=obj_id)

    def create_index(self, keys, unique=False, where_exist=False):
        """Create an index on the metadata.  Takes either a single key or list of (key, direction)
         pairs

         :param keys: the key or keys to create the index on
         :param unique: if True, create a uniqueness constraint on this index
         :param where_exist: if True, only apply this index on documents that contain the key(s)
         """
        self._archive.meta_create_index(keys, unique=unique, where_exist=where_exist)



import contextlib
import functools
import uuid

import deprecation

from . import base_savable
from . import tracking
from . import version

__all__ = ('Process',)


@deprecation.deprecated(deprecated_in="0.14.5",
                        removed_in="0.16",
                        current_version=version.__version__,
                        details="Use mincepy.track() instead")
def track(func):

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.running():
            return func(self, *args, **kwargs)

    return wrapper


class Process(base_savable.SimpleSavable):
    TYPE_ID = uuid.UUID('bcf03171-a1f1-49c7-b890-b7f9d9f9e5a2')
    STACK = []
    ATTRS = '_name', '_running'

    def __init__(self, name: str):
        super(Process, self).__init__()
        self._name = name
        self._running = 0

    def __eq__(self, other):
        if not isinstance(other, Process):
            return False

        return self.name == other.name

    @property
    def is_running(self):
        return self._running != 0

    @property
    def name(self) -> str:
        return self._name

    @contextlib.contextmanager
    def running(self):
        try:
            self._running += 1
            with tracking.track(self):
                yield
        finally:
            self._running -= 1



from typing import Optional, Sequence, Union, Iterable, Mapping, Iterator, Dict, Tuple
import uuid

import bson
import gridfs
import networkx
import pymongo
import pymongo.uri_parser
import pymongo.database
import pymongo.errors

import mincepy
import mincepy.records
from mincepy import operations
from mincepy import q
from mincepy import exceptions

from . import bulk
from . import migrate
from . import migrations
from . import files
from . import db
from . import references
from . import queries

__all__ = ('MongoArchive', 'connect')

DEFAULT_REFERENCES_COLLECTION = 'references'

scalar_query_spec = mincepy.archives.scalar_query_spec  # pylint: disable=invalid-name


class ObjectIdHelper(mincepy.TypeHelper):
    TYPE = bson.ObjectId
    TYPE_ID = uuid.UUID('bdde0765-36d2-4f06-bb8b-536a429f32ab')

    def yield_hashables(self, obj, hasher):
        yield obj.binary

    def eq(self, one, other) -> bool:
        return one.__eq__(other)

    def save_instance_state(self, obj, _depositor):
        return obj

    def load_instance_state(self, obj, saved_state, _depositor):
        return obj.__init__(saved_state)


class MongoArchive(mincepy.BaseArchive[bson.ObjectId]):
    """MongoDB implementation of the mincepy archive"""

    # pylint: disable=too-many-public-methods

    ID_TYPE = bson.ObjectId
    SnapshotId = mincepy.SnapshotId[bson.ObjectId]

    DATA_COLLECTION = 'data'
    HISTORY_COLLECTION = 'history'
    META_COLLECTION = 'meta'

    @classmethod
    def get_types(cls) -> Sequence:
        return ObjectIdHelper(), files.GridFsFile

    def __init__(self, database: pymongo.database.Database):
        self._database = database
        migrate.ensure_up_to_date(database, migrations.LATEST)

        self._data_collection = database[self.DATA_COLLECTION]
        self._history_collection = database[self.HISTORY_COLLECTION]
        self._meta_collection = database[self.META_COLLECTION]
        self._file_bucket = gridfs.GridFSBucket(database)
        self._refman = references.ReferenceManager(database[DEFAULT_REFERENCES_COLLECTION],
                                                   self._data_collection, self._history_collection)
        self._create_indices()

    @property
    def database(self) -> pymongo.database.Database:
        return self._database

    @property
    def data_collection(self) -> pymongo.database.Collection:
        return self._data_collection

    def _create_indices(self):
        # Create all the necessary indexes
        self._data_collection.create_index(db.OBJ_ID)
        self._history_collection.create_index([(db.OBJ_ID, pymongo.ASCENDING),
                                               (db.VERSION, pymongo.ASCENDING)])

    def create_archive_id(self):  # pylint: disable=no-self-use
        return bson.ObjectId()

    def construct_archive_id(self, value) -> bson.ObjectId:  # pylint: disable=no-self-use
        if not isinstance(value, str):
            raise TypeError("Cannot construct an ObjectID from a '{}'".format(type(value)))
        try:
            return bson.ObjectId(value)
        except bson.errors.InvalidId as exc:
            raise ValueError(str(exc))

    def create_file(self, filename: str = None, encoding: str = None):
        return files.GridFsFile(self._file_bucket, filename, encoding)

    def get_gridfs_bucket(self) -> gridfs.GridFSBucket:
        return self._file_bucket

    def bulk_write(self, ops: Sequence[operations.Operation]):
        # First, convert these to corresponding mongo bulk operations.  Because of the way we split
        # objects into 'data' and 'history' we have to perform these operations on both
        data_ops = []
        history_ops = []

        for data_op, history_op in map(bulk.to_mongo_op, ops):
            data_ops.append(data_op)
            history_ops.append(history_op)

        try:
            # First perform the data operations
            self._data_collection.bulk_write(data_ops, ordered=True)
            # Then the history operations
            self._history_collection.bulk_write(history_ops, ordered=True)
        except pymongo.errors.BulkWriteError as exc:
            write_errors = exc.details['writeErrors']
            if write_errors:
                raise mincepy.ModificationError(
                    "You're trying to rewrite history, that's not allowed!")
            raise  # Otherwise just raise what we got

        self._refman.invalidate(obj_ids=[op.obj_id for op in ops],
                                snapshot_ids=[op.snapshot_id for op in ops])

    def load(self, snapshot_id: mincepy.SnapshotId) -> mincepy.DataRecord:
        if not isinstance(snapshot_id, mincepy.SnapshotId):
            raise TypeError(snapshot_id)

        results = tuple(
            self._history_collection.find({
                db.OBJ_ID: snapshot_id.obj_id,
                db.VERSION: snapshot_id.version
            }))
        if not results:
            raise mincepy.NotFound("Snapshot id '{}' not found".format(snapshot_id))
        return db.to_record(results[0])

    def get_snapshot_ids(self, obj_id: bson.ObjectId):
        results = self._history_collection.find({db.OBJ_ID: obj_id},
                                                projection={
                                                    db.OBJ_ID: 1,
                                                    db.VERSION: 1
                                                },
                                                sort=[(db.VERSION, pymongo.ASCENDING)])
        if not results:
            return []

        return list(map(db.sid_from_dict, results))

    # region Meta

    def meta_get(self, obj_id: Union[bson.ObjectId, Iterable[bson.ObjectId]]):
        # Single obj id
        if not isinstance(obj_id, bson.ObjectId):
            raise TypeError("Must pass an ObjectId, got {}".format(obj_id))
        found = self._meta_collection.find_one({'_id': obj_id})
        if found is None:
            return found
        found.pop('_id')
        return found

    def meta_get_many(self, obj_ids: Iterable[bson.ObjectId]) -> Dict[bson.ObjectId, dict]:
        # Find multiple
        for obj_id in obj_ids:
            if not isinstance(obj_id, bson.ObjectId):
                raise TypeError("Must pass an ObjectId, got {}".format(obj_id))

        cur = self._meta_collection.find({'_id': q.in_(*obj_ids)})
        results = {oid: None for oid in obj_ids}
        for found in cur:
            results[found.pop('_id')] = found

        return results

    def meta_set(self, obj_id, meta):
        if meta:
            try:
                found = self._meta_collection.replace_one({'_id': obj_id}, meta, upsert=True)
            except pymongo.errors.DuplicateKeyError as exc:
                raise mincepy.DuplicateKeyError(str(exc))
            else:
                if not found:
                    raise mincepy.NotFound("No record with snapshot id '{}' found".format(obj_id))
        else:
            # Just remove the meta entry outright
            self._meta_collection.delete_one({'_id': obj_id})

    def meta_set_many(self, metas: Mapping[bson.ObjectId, Optional[dict]]):
        ops = []
        for obj_id, meta in metas.items():
            if meta:
                operation = pymongo.operations.ReplaceOne({'_id': obj_id}, meta, upsert=True)
            else:
                operation = pymongo.operations.DeleteOne({'_id': obj_id})
            ops.append(operation)

        try:
            self._meta_collection.bulk_write(ops, ordered=True)
        except pymongo.errors.BulkWriteError as exc:
            # This is a rather complicated way to get the error - mongo doesn't seem to document
            # error codes, absolute madness.
            if exc.code == 65:
                write_errors = exc.details.get('writeErrors', {})
                if write_errors:
                    # There should only be one as it is an ordered write
                    error = write_errors[0]
                    if error.get('code') == 11000:
                        raise mincepy.DuplicateKeyError(error.get('errmsg'))
            raise

    def meta_update(self, obj_id, meta: Mapping):
        if meta.get('_id', obj_id) != obj_id:
            raise ValueError("Cannot use the key _id, in metadata: it is reserved")

        try:
            self._meta_collection.update_one({'_id': obj_id}, {'$set': meta}, upsert=True)
        except pymongo.errors.DuplicateKeyError as exc:
            raise mincepy.DuplicateKeyError(str(exc))

    def meta_find(
        self,
        filter: dict = None,  # pylint: disable=redefined-builtin
        obj_id: Union[bson.ObjectId, Iterable[bson.ObjectId], Dict] = None
    ) -> Iterator[Tuple[bson.ObjectId, Dict]]:
        match = dict(filter or {})
        if obj_id is not None:
            match['_id'] = scalar_query_spec(obj_id)

        for meta in self._meta_collection.find(match):
            oid = meta.pop('_id')
            yield self.MetaEntry(oid, meta)

    def meta_distinct(
        self,
        key: str,
        filter: dict = None,  # pylint: disable=redefined-builtin
        obj_id: Union[bson.ObjectId, Iterable[bson.ObjectId], Mapping] = None
    ) -> 'Iterator':
        match = dict(filter or {})
        if obj_id is not None:
            match['_id'] = scalar_query_spec(obj_id)

        yield from self._meta_collection.distinct(key, match)

    def meta_create_index(self, keys, unique=True, where_exist=False):
        kwargs = {}
        if where_exist:
            if not isinstance(keys, str) and isinstance(keys, Iterable):
                key_names = tuple(entry[0] for entry in keys)
            else:
                key_names = (keys,)
            kwargs['partialFilterExpression'] = \
                q.and_(*tuple(q.exists_(name) for name in key_names))
        self._meta_collection.create_index(keys, unique=unique, **kwargs)

    # endregion

    # pylint: disable=too-many-arguments
    def find(self,
             obj_id: Union[bson.ObjectId, Iterable[bson.ObjectId], Dict] = None,
             type_id: Union[bson.ObjectId, Iterable[bson.ObjectId], Dict] = None,
             _created_by=None,
             _copied_from=None,
             version=None,
             state=None,
             state_types=None,
             snapshot_hash=None,
             meta: dict = None,
             extras: dict = None,
             limit=0,
             sort=None,
             skip=0):
        pipeline = self._get_pipeline(obj_id=obj_id,
                                      type_id=type_id,
                                      _created_by=_created_by,
                                      _copied_from=_copied_from,
                                      version=version,
                                      state=state,
                                      state_types=state_types,
                                      snapshot_hash=snapshot_hash,
                                      meta=meta,
                                      extras=extras)

        if skip:
            pipeline.append({'$skip': skip})

        if limit:
            pipeline.append({'$limit': limit})

        if sort:
            if not isinstance(sort, dict):
                sort_dict = {sort: 1}
            else:
                sort_dict = sort
            sort_dict = db.remap(sort_dict)
            pipeline.append({'$sort': sort_dict})

        if version == -1:
            coll = self._data_collection
        else:
            coll = self._history_collection

        results = coll.aggregate(pipeline, allowDiskUse=True)

        for result in results:
            yield db.to_record(result)

    def count(self,
              obj_id: Optional[bson.ObjectId] = None,
              type_id=None,
              _created_by=None,
              _copied_from=None,
              version=-1,
              state=None,
              snapshot_hash=None,
              meta=None,
              limit=0):

        pipeline = self._get_pipeline(obj_id=obj_id,
                                      type_id=type_id,
                                      _created_by=_created_by,
                                      _copied_from=_copied_from,
                                      version=version,
                                      state=state,
                                      snapshot_hash=snapshot_hash,
                                      meta=meta)

        if version == -1:
            coll = self._data_collection
        else:
            coll = self._history_collection

        if meta:
            pipeline.extend(
                queries.pipeline_match_metadata(meta, self._meta_collection.name, db.OBJ_ID))

        if limit:
            pipeline.append({'$limit': limit})

        pipeline.append({'$count': "total"})
        result = next(coll.aggregate(pipeline))

        return result['total']

    def get_snapshot_ref_graph(self,
                               *snapshot_ids: SnapshotId,
                               direction=mincepy.OUTGOING,
                               max_dist: int = None) -> Iterator[networkx.DiGraph]:
        return self._refman.get_snapshot_ref_graph(snapshot_ids,
                                                   direction=direction,
                                                   max_dist=max_dist)

    def get_obj_ref_graph(self,
                          *obj_ids: bson.ObjectId,
                          direction=mincepy.OUTGOING,
                          max_dist: int = None) -> Iterator[networkx.DiGraph]:
        return self._refman.get_obj_ref_graphs(obj_ids, direction=direction, max_dist=max_dist)

    def _get_pipeline(self,
                      obj_id: Union[bson.ObjectId, Iterable[bson.ObjectId]] = None,
                      type_id=None,
                      _created_by=None,
                      _copied_from=None,
                      version=None,
                      state=None,
                      state_types=None,
                      snapshot_hash=None,
                      meta: dict = None,
                      extras: dict = None):
        """Get a pipeline that would perform the given search.  Can be used directly in an aggregate
         call"""
        pipeline = []

        query = queries.QueryBuilder()

        if obj_id is not None:
            query.and_({db.OBJ_ID: scalar_query_spec(obj_id)})

        if version is not None and version != -1:
            query.and_({db.VERSION: version})

        if type_id is not None:
            query.and_({db.TYPE_ID: scalar_query_spec(type_id)})

        if state is not None:
            query.and_(*queries.flatten_filter(db.STATE, state))

        if state_types is not None:
            query.and_(*queries.flatten_filter(db.STATE_TYPES, state_types))

        if snapshot_hash is not None:
            query.and_({db.SNAPSHOT_HASH: scalar_query_spec(snapshot_hash)})

        if extras:
            query.and_(*queries.flatten_filter(db.EXTRAS, extras))

        mfilter = query.build()
        if mfilter:
            pipeline.append({'$match': mfilter})

        if meta:
            pipeline.extend(
                queries.pipeline_match_metadata(meta, self._meta_collection.name, db.OBJ_ID))

        return pipeline


def connect(uri: str) -> MongoArchive:
    # URI Format is:
    # mongodb://[username:password@]host1[:port1][,...hostN[:portN]][/[database][?options]]
    parsed = pymongo.uri_parser.parse_uri(uri)
    if not parsed.get('database', None):
        raise ValueError("Failed to supply database on MongoDB uri: {}".format(uri))

    try:
        client = pymongo.MongoClient(uri)
    except pymongo.errors.ServerSelectionTimeoutError as exc:
        raise exceptions.ConnectionError(str(exc))

    database = client[parsed['database']]
    return MongoArchive(database)



import tempfile
import uuid

import bson
import gridfs

import mincepy

__all__ = ('GridFsFile',)


class GridFsFile(mincepy.BaseFile):
    TYPE_ID = uuid.UUID('3bf3c24e-f6c8-4f70-956f-bdecd7aed091')
    ATTRS = '_persistent_id', '_file_id'

    def __init__(self,
                 file_bucket: gridfs.GridFSBucket,
                 filename: str = None,
                 encoding: str = None):
        super().__init__(filename, encoding)
        self._file_store = file_bucket
        self._file_id = None
        self._persistent_id = bson.ObjectId()
        self._buffer_file = _create_buffer_file()

    def open(self, mode='r', **kwargs):
        self._ensure_buffer()
        if 'b' not in mode:
            kwargs.setdefault('encoding', self.encoding)
        return open(self._buffer_file, mode, **kwargs)

    def save_instance_state(self, saver: mincepy.Saver):
        filename = self.filename or ""
        with open(self._buffer_file, 'rb') as fstream:
            self._file_id = self._file_store.upload_from_stream(filename, fstream)

        return super().save_instance_state(saver)

    def load_instance_state(self, saved_state, loader: mincepy.Loader):
        super().load_instance_state(saved_state, loader)
        self._file_store = loader.get_archive().get_gridfs_bucket()  # type: gridfs.GridFSBucket
        # Don't copy the file over now, do it lazily when the file is first opened
        self._buffer_file = None

    def _ensure_buffer(self):
        if self._buffer_file is None:
            if self._file_id is not None:
                self._update_buffer()
            else:
                _create_buffer_file()

    def _update_buffer(self):
        self._buffer_file = _create_buffer_file()
        with open(self._buffer_file, 'wb') as fstream:
            self._file_store.download_to_stream(self._file_id, fstream)


def _create_buffer_file():
    tmp_file = tempfile.NamedTemporaryFile(delete=False)
    tmp_path = tmp_file.name
    tmp_file.close()
    return tmp_path



from . import comparators


def get_default_equators():
    return (comparators.SequenceEquator(), comparators.BytesEquator(), comparators.StrEquator(),
            comparators.SetEquator(), comparators.OrderedDictEquator(),
            comparators.MappingEquator(), comparators.ComplexEquator(), comparators.RealEquator(),
            comparators.IntegerEquator(), comparators.BoolEquator(), comparators.NoneEquator(),
            comparators.TupleEquator(), comparators.UuidEquator())



import io
import os
import shutil

import pytest

import mincepy


def test_file_basics(historian: mincepy.Historian):
    encoding = 'utf-8'
    INITIAL_DATA = 'hello there'
    file = historian.create_file(encoding)
    with file.open('w') as stream:
        stream.write(INITIAL_DATA)

    file_id = historian.save(file)
    del file

    loaded = historian.load(file_id)
    with loaded.open('r') as file:
        buffer = io.StringIO()
        shutil.copyfileobj(file, buffer)
        assert buffer.getvalue() == INITIAL_DATA


def test_file_changing(tmp_path, historian: mincepy.Historian):
    encoding = 'utf-8'
    INITIAL_DATA = "Initial string"
    mince_file = historian.create_file(encoding=encoding)

    with mince_file.open('w') as file:
        file.write(INITIAL_DATA)

    historian.save(mince_file)

    # Now let's append to the file
    NEW_DATA = "Second string"
    with mince_file.open('a') as file:
        file.write(NEW_DATA)

    historian.save(mince_file)
    history = historian.history(mince_file)
    assert len(history) == 2

    with history[0].obj.open() as file:
        buffer = io.StringIO()
        shutil.copyfileobj(file, buffer)
        assert INITIAL_DATA == buffer.getvalue()

    with history[1].obj.open() as file:
        buffer = io.StringIO()
        shutil.copyfileobj(file, buffer)
        assert INITIAL_DATA + NEW_DATA == buffer.getvalue()


def test_nested_files_in_list(historian: mincepy.Historian):
    file = historian.create_file()
    my_list = mincepy.builtins.List()
    my_list.append(file)

    list_id = historian.save(my_list)
    del my_list

    loaded = historian.load(list_id)
    assert len(loaded) == 1
    assert loaded[0].filename == None


def test_nested_files_in_dict(historian: mincepy.Historian):
    file = historian.create_file()
    my_dict = mincepy.builtins.Dict()
    my_dict['file'] = file

    list_id = historian.save(my_dict)
    del my_dict

    loaded = historian.load(list_id)
    assert len(loaded) == 1
    assert loaded['file'].filename == None


def test_nested_files_in_list_mutating(tmp_path, historian: mincepy.Historian):
    encoding = 'utf-8'
    INITIAL_DATA = "First string".encode(encoding)
    my_file = historian.create_file()
    with my_file.open('wb') as file:
        file.write(INITIAL_DATA)

    my_list = mincepy.builtins.List()
    my_list.append(my_file)

    list_id = historian.save(my_list)

    # Now let's append to the file
    NEW_DATA = "Second string".encode(encoding)
    with my_file.open('ab') as file:
        file.write(NEW_DATA)

    # Save the list again
    historian.save(my_list)
    del my_list

    loaded = historian.load(list_id)
    with loaded[0].open('rb') as contents:
        buffer = io.BytesIO()
        shutil.copyfileobj(contents, buffer)
        assert buffer.getvalue() == INITIAL_DATA + NEW_DATA



from click.testing import CliRunner

import mincepy
import mincepy.cli.main
from ..common import CarV1, CarV2, StoreByValue, StoreByRef


def test_simple_migrate(historian: mincepy.Historian, archive_uri):
    car = CarV1('white', 'lada')
    car.save()
    by_val = StoreByValue(car)
    by_val_id = by_val.save()

    # Register a new version of the car
    historian.register_type(CarV2)

    # Now both car and by_val should need migration (because by_val stores a car)
    migratable = tuple(historian.migrations.find_migratable_records())
    assert len(migratable) == 2

    # Now migrate
    runner = CliRunner()
    result = runner.invoke(mincepy.cli.main.migrate, ["--yes", archive_uri],
                           obj={'helpers': [CarV2, StoreByValue]})
    assert result.exit_code == 0

    migratable = tuple(historian.migrations.find_migratable_records())
    assert len(migratable) == 0

    # Now register a new version of StoreByVal
    historian.register_type(StoreByRef)

    # There should still be the same to migratables as before
    migratable = tuple(historian.migrations.find_migratable_records())
    assert len(migratable) == 1
    ids = [record.obj_id for record in migratable]
    assert by_val_id in ids

    # Now migrate
    runner = CliRunner()
    result = runner.invoke(mincepy.cli.main.migrate, ["--yes", archive_uri],
                           obj={'helpers': [CarV2, StoreByRef]})
    assert result.exit_code == 0

    migratable = tuple(historian.migrations.find_migratable_records())
    assert len(migratable) == 0



import pytest

import mincepy
from mincepy import testing

# pylint: disable=invalid-name


def test_basic_save_process(historian: mincepy.Historian):
    proc = mincepy.Process('test_basic_save')
    pid = historian.save(proc)
    with proc.running():
        car = testing.Car('nissan', 'white')
        car_id = historian.save(car)
    assert historian.created_by(car) == pid

    second_car = testing.Car('ford')
    historian.save(second_car)
    assert historian.created_by(second_car) is None

    # Now check we can get the creator id from the object id
    del car
    assert historian.created_by(car_id) == pid


def test_save_after_creation(historian: mincepy.Historian):
    """
    Test saving an object that was created inside a process context but then saved
    outside it.  The creator should still be correctly set
    """
    proc = mincepy.Process('test_delayed_save')
    proc.save()
    with proc.running():
        # Create the car
        car = testing.Car('nissan', 'white')

    # Save it
    historian.save(car)
    created_in = historian.created_by(car)
    assert created_in is not None
    assert created_in == historian.get_current_record(proc).obj_id


def test_process_nested_running(historian: mincepy.Historian):
    proc = mincepy.Process('test_nested_exception')
    with proc.running():
        with proc.running():
            pass
        assert proc.is_running
    historian.save(proc)

    # Now check that nested exceptions are correctly handled
    with pytest.raises(TypeError):
        with proc.running():
            with pytest.raises(RuntimeError):
                with proc.running():
                    raise RuntimeError('Failed yo')
            assert proc.is_running
            raise TypeError("New error")
        assert proc.is_running
    proc_id = historian.save(proc)
    del proc

    loaded = historian.load(proc_id)
    assert not loaded.is_running


def test_saving_while_running(historian: mincepy.Historian):
    proc = mincepy.Process('test_nested_exception')
    with proc.running():
        historian.save(proc)


def test_saving_creator_that_owns_child(historian: mincepy.Historian):

    class TestProc(mincepy.Process):
        ATTRS = ('child',)

        def __init__(self):
            super().__init__('test_proc')
            self.child = None

    test_proc = TestProc()
    with test_proc.running():
        test_proc.child = mincepy.ObjRef(testing.Car())
        historian.save(test_proc)


def test_process_track(historian: mincepy.Historian):

    class TestProc(mincepy.Process):

        @mincepy.track
        def execute(self):  # pylint: disable=no-self-use
            return mincepy.builtins.RefList([testing.Car()])

    proc = TestProc('test process')
    proc.save()
    car_list = proc.execute()
    historian.save(car_list)
    proc_id = historian.get_obj_id(proc)

    assert proc_id is not None
    assert historian.get_current_record(car_list).created_by == proc_id
    assert historian.get_current_record(car_list[0]).created_by == proc_id



from typing import Generic, TypeVar, Set, overload

import networkx
from networkx.algorithms import dag

from mincepy import archives
from mincepy import records
from mincepy import operations
from mincepy import transactions

__all__ = ('References',)

IdT = TypeVar('IdT')  # The archive ID type


class References(Generic[IdT]):
    """
    A class that can provide reference graph information about objects stored in the archive.

    .. note::
        It is deliberately not possible to pass an object directly to methods in this class as what
        is returned is reference information from the archive and _not_ reference information about
        the in-memory python object.
    """

    def __init__(self, historian):
        self._historian = historian
        self._archive = historian.archive  # type: archives.Archive

    SnapshotId = records.SnapshotId[IdT]

    @overload
    def references(self, identifier: IdT) -> Set[IdT]:  # pylint: disable=no-self-use
        ...

    @overload
    def references(self, identifier: 'SnapshotId') -> 'Set[SnapshotId]':  # pylint: disable=no-self-use
        ...

    def references(self, identifier):
        """Get the ids of the objects referred to by the passed identifier."""
        if isinstance(identifier, records.SnapshotId):
            graph = self.get_snapshot_ref_graph(identifier, max_dist=1)
        elif isinstance(identifier, self._archive.get_id_type()):
            graph = self.get_obj_ref_graph(identifier, max_dist=1)
        else:
            raise TypeError(identifier)

        return set(edge[1] for edge in graph.edges)

    @overload
    def referenced_by(self, identifier: IdT) -> 'Set[IdT]':  # pylint: disable=no-self-use
        ...

    @overload
    def referenced_by(self, identifier: 'SnapshotId') -> 'Set[SnapshotId]':  # pylint: disable=no-self-use
        ...

    def referenced_by(self, identifier):
        """Get the ids of the objects that refer to the passed object"""
        if isinstance(identifier, records.SnapshotId):
            graph = self.get_snapshot_ref_graph(identifier, direction=archives.INCOMING, max_dist=1)
        elif isinstance(identifier, self._archive.get_id_type()):
            graph = self.get_obj_ref_graph(identifier, direction=archives.INCOMING, max_dist=1)
        else:
            raise TypeError(identifier)

        return set(edge[0] for edge in graph.edges)

    def get_snapshot_ref_graph(self,
                               *snapshot_ids: SnapshotId,
                               direction=archives.OUTGOING,
                               max_dist: int = None) -> networkx.DiGraph:

        return self._archive.get_snapshot_ref_graph(*snapshot_ids,
                                                    direction=direction,
                                                    max_dist=max_dist)

    def get_obj_ref_graph(self,
                          *obj_ids: IdT,
                          direction=archives.OUTGOING,
                          max_dist: int = None) -> networkx.DiGraph:
        obj_ids = set(obj_ids)
        graph = self._archive.get_obj_ref_graph(*obj_ids, direction=direction, max_dist=max_dist)

        # If there is a transaction then we should fix up the graph to contain information from that
        # too
        trans = self._historian.current_transaction()  # type: transactions.Transaction
        if trans is not None:
            _update_from_transaction(graph, trans)

            # Now cull all the nodes not reachable from the nodes of interest

            # Now, get the subgraph we're interested in
            reachable = set()
            for obj_id in obj_ids:
                if direction == archives.OUTGOING:
                    reachable.update(dag.descendants(graph, obj_id))
                else:
                    reachable.update(dag.ancestors(graph, obj_id))

            # Remove all non-reachable nodes except obj_ids as these can stay even if they have no
            # edges
            graph.remove_nodes_from(set(graph.nodes) - obj_ids - reachable)

        return graph


def _update_from_transaction(graph: networkx.DiGraph, transaction: transactions.Transaction):
    """Given a transaction update the reference graph to reflect the insertion of any new records"""
    for op in transaction.staged:  # pylint: disable=invalid-name
        if isinstance(op, operations.Insert):
            # Modify the graph to reflect the insertion
            obj_id = op.obj_id
            if obj_id in graph.nodes:
                # Incoming edges (things referencing this object) can stay, as they haven't
                # changed but outgoing edges may have
                out_edges = tuple(graph.out_edges(obj_id))
                graph.remove_edges_from(out_edges)

            # And add in the current ones
            for refs in op.record.get_references():
                graph.add_edge(obj_id, refs[1].obj_id)



import pytest

import mincepy
from mincepy.testing import Person, Car, Garage

# pylint: disable=invalid-name


def test_references_wrong_type(historian: mincepy.Historian):
    with pytest.raises(TypeError):
        historian.references.references(1234)


def test_referenced_by_wrong_type(historian: mincepy.Historian):
    with pytest.raises(TypeError):
        historian.references.referenced_by(1234)


def test_references_simple(historian: mincepy.Historian):
    address_book = mincepy.RefList()
    for i in range(3):
        address_book.append(Person(name='test', age=i))
    address_book.save()

    refs = historian.references.references(address_book.obj_id)
    assert len(refs) == len(address_book)
    assert not set(person.obj_id for person in address_book) - set(refs)

    refs = historian.references.referenced_by(address_book[0].obj_id)
    assert len(refs) == 1
    assert address_book.obj_id in refs


def test_remove_references_in_transaction(historian: mincepy.Historian):
    car = Car()
    garage = Garage(mincepy.ObjRef(car))
    gid = historian.save(garage)

    graph = historian.references.get_obj_ref_graph(gid)
    assert len(graph.nodes) == 2
    assert len(graph.edges) == 1
    assert (gid, car.obj_id) in graph.edges

    reffed_by = historian.references.referenced_by(car.obj_id)
    assert len(reffed_by) == 1
    assert gid in reffed_by

    # Now, change the reference in a transaction
    with historian.transaction():
        garage.car = None
        garage.save()
        # Still in a transaction, check the references are correct, i.e. None
        graph = historian.references.get_obj_ref_graph(gid)

        assert len(graph.nodes) == 1
        assert gid in graph.nodes
        assert len(graph.edges) == 0

        reffed_by = historian.references.referenced_by(car.obj_id)
        assert len(reffed_by) == 0

    assert len(historian.references.referenced_by(car.obj_id)) == 0
    assert len(historian.references.references(gid)) == 0


def test_add_references_in_transaction(historian: mincepy.Historian):
    car = Car()
    garage = mincepy.RefList()
    garage.append(car)
    historian.save(garage)

    refs = historian.references.references(garage.obj_id)
    assert len(refs) == 1
    assert car.obj_id in refs

    # Now add a car
    with historian.transaction():
        car2 = Car()
        garage.append(car2)
        garage.save()

        # Still in a transaction, check the references are correct
        refs = historian.references.references(garage.obj_id)
        assert len(refs) == 2
        assert car.obj_id in refs
        assert car2.obj_id in refs

        reffed_by = historian.references.referenced_by(car2.obj_id)
        assert len(reffed_by) == 1
        assert garage.obj_id in reffed_by

    refs = historian.references.references(garage.obj_id)
    assert len(refs) == 2
    assert car.obj_id in refs
    assert car2.obj_id in refs

    reffed_by = historian.references.referenced_by(car2.obj_id)
    assert len(reffed_by) == 1
    assert garage.obj_id in reffed_by


def test_snapshot_references(historian: mincepy.Historian):
    address_book = mincepy.RefList()
    for i in range(3):
        address_book.append(Person(name='test', age=i))
    address_book.save()
    sid = historian.get_snapshot_id(address_book)

    refs = historian.references.references(sid)
    assert len(refs) == len(address_book)
    assert not set(historian.get_snapshot_id(person) for person in address_book) - set(refs)


def test_snapshot_referenced_by(historian: mincepy.Historian):
    address_book = mincepy.RefList()
    for i in range(3):
        address_book.append(Person(name='test', age=i))
    address_book.save()
    sid = historian.get_snapshot_id(address_book)

    refs = historian.references.referenced_by(historian.get_snapshot_id(address_book[0]))
    assert len(refs) == 1
    assert sid in refs



import copy as python_copy
import functools
from typing import Callable

from . import records
from . import staging

__all__ = 'track', 'copy', 'deepcopy', 'mark_as_copy'


class TrackStack:
    """Stack to keep track of functions being called"""

    _stack = []

    def __init__(self):
        raise RuntimeError("Cannot be instantiated")

    @classmethod
    def peek(cls):
        if not cls._stack:
            return None
        return cls._stack[-1]

    @classmethod
    def push(cls, obj):
        """Push an object onto the stack"""
        cls._stack.append(obj)

    @classmethod
    def pop(cls, obj):
        """Pop an object off of the stack"""
        if cls._stack[-1] != obj:
            raise RuntimeError("Someone has corrupted the process stack!\n"
                               "Expected to find '{}' on top but found '{}'".format(
                                   obj, cls._stack[-1]))

        cls._stack.pop()


class TrackContext:
    """Context manager that pushes and pops the object to/from the track stack"""

    def __init__(self, obj):
        self._obj = obj

    def __enter__(self):
        TrackStack.push(self._obj)

    def __exit__(self, exc_type, exc_val, exc_tb):
        TrackStack.pop(self._obj)


def track(obj_or_fn):
    """Allows object creation to be tracked.  When an object is created within this context, the
    creator of the object will be saved in the database record.

    This can be used either as a decorator to a class method, in which case the object instance will
    be the creator.  Or it can be used as a context in which case the creator should be passed as
    the argument.
    """
    if isinstance(obj_or_fn, Callable):
        # We're acting as a decorator
        @functools.wraps(obj_or_fn)
        def wrapper(self, *args, **kwargs):
            with TrackContext(self):
                return obj_or_fn(self, *args, **kwargs)

        return wrapper

    # We're acting as a context
    return TrackContext(obj_or_fn)


def obj_created(obj):
    creator = TrackStack.peek()
    if creator is not None:
        staging.get_info(obj)[records.ExtraKeys.CREATED_BY] = creator


def mark_as_copy(obj, copied_from):
    staging.get_info(obj)[records.ExtraKeys.COPIED_FROM] = copied_from


def copy(obj):
    """Create a shallow copy of the object.  Using this method allows the historian to inject
    information about where the object was copied from into the record if saved."""
    obj_copy = python_copy.copy(obj)
    mark_as_copy(obj_copy, obj)
    return obj_copy


def deepcopy(obj):
    """Create a shallow copy of the object.  Using this method allows the historian to inject
    information about where the object was copied from into the record if saved."""
    obj_copy = python_copy.deepcopy(obj)
    mark_as_copy(obj_copy, obj)
    return obj_copy



from . import qops as q
from .archives import *
from .builtins import *
from .base_savable import *
from .archive_factory import *
from .comparators import *
from .depositors import *
from .exceptions import *  # pylint: disable=redefined-builtin
from .helpers import *
from .historians import *
from .history import *
from .migrations import *
from .process import *
from .records import *
from .refs import *
from .tracking import *
from .types import *
from .version import *
from . import builtins
from . import common_helpers
from . import history
from . import mongo
from . import operations
from . import process
from . import utils

_ADDITIONAL = 'mongo', 'builtins', 'common_helpers', 'utils', 'q', 'operations'

__all__ = (archives.__all__ + depositors.__all__ + exceptions.__all__ + historians.__all__ +
           process.__all__ + types.__all__ + helpers.__all__ + version.__all__ + history.__all__ +
           archive_factory.__all__ + refs.__all__ + records.__all__ + base_savable.__all__ +
           builtins.__all__ + migrations.__all__ + tracking.__all__ + _ADDITIONAL)



import mincepy
from mincepy.testing import Car, Garage, Cycle


def test_get_snapshot_graph_simple(historian: mincepy.Historian):
    car = Car()
    garage = Garage(mincepy.ObjRef(car))
    garage.save()
    garage_sid = historian.get_snapshot_id(garage)

    garage_graph = historian.archive.get_snapshot_ref_graph(garage_sid)
    assert len(garage_graph.edges) == 1
    assert (garage_sid, historian.get_snapshot_id(car)) in garage_graph.edges


def test_get_snapshot_self_cycle(historian: mincepy.Historian):
    node = Cycle()
    node.ref = node  # Cycle complete
    historian.save_one(node)

    node_ref = historian.get_snapshot_id(node)
    node_graph = historian.archive.get_snapshot_ref_graph(node_ref)
    assert len(node_graph.edges) == 1
    assert (node_ref, node_ref) in node_graph.edges


def test_get_snapshot_graph_cycle(historian: mincepy.Historian):
    node1 = Cycle()
    node2 = Cycle(node1)
    node3 = Cycle(node2)
    node1.ref = node3  # Cycle complete

    historian.save(node1, node2, node3)

    node1ref = historian.get_snapshot_id(node1)
    node2ref = historian.get_snapshot_id(node2)
    node3ref = historian.get_snapshot_id(node3)

    graph = historian.archive.get_snapshot_ref_graph(node1ref)
    assert len(graph.edges) == 3

    # Created the edges to check
    n13 = (node1ref, node3ref)
    n21 = (node2ref, node1ref)
    n32 = (node3ref, node2ref)

    assert n13 in graph.edges
    assert n21 in graph.edges
    assert n32 in graph.edges


def test_get_snapshot_graph_twice(historian: mincepy.Historian):
    """Check for a bug that arise when asking for references twice"""
    car = Car()
    garage = Garage(mincepy.ObjRef(car))
    garage.save()
    garage_sid = historian.get_snapshot_id(garage)

    def make_checks(graph):
        assert len(graph.edges) == 1
        assert (garage_sid, historian.get_snapshot_id(car)) in graph.edges

    ref_graphs = historian.archive.get_snapshot_ref_graph(garage_sid)
    make_checks(ref_graphs)

    # Check again
    ref_graphs = historian.archive.get_snapshot_ref_graph(garage_sid)
    make_checks(ref_graphs)


def test_get_object_graph(historian: mincepy.Historian):
    """Try getting the reference graph for live objects"""
    car = Car()
    garage = Garage(mincepy.ObjRef(car))
    gid = garage.save()

    garage_graph = historian.archive.get_obj_ref_graph(gid)
    assert len(garage_graph.edges) == 1
    assert (gid, car.obj_id) in garage_graph.edges


def test_get_obj_graph_current(historian: mincepy.Historian):
    """Test that when changing a reference the reference graph is returned correctly"""
    car = Car()
    garage = Garage(mincepy.ObjRef(car))
    gid = garage.save()

    garage_graph = historian.archive.get_obj_ref_graph(gid)
    assert len(garage_graph.edges) == 1
    assert (gid, car.obj_id) in garage_graph.edges

    # Now, modify the garage
    car2 = Car()
    garage.car = mincepy.ObjRef(car2)
    garage.save()

    # Check that the reference graph is correct
    garage_graph = historian.archive.get_obj_ref_graph(gid)
    assert len(garage_graph.edges) == 1
    assert (gid, car2.obj_id) in garage_graph.edges

    # Finally, set the reference to None
    garage.car = mincepy.ObjRef()
    garage.save()

    # Check that the reference graph is correct
    garage_graph = historian.archive.get_obj_ref_graph(gid)
    assert len(garage_graph.edges) == 0
    assert len(garage_graph.nodes) == 1


def test_get_obj_referencing_simple(historian: mincepy.Historian):
    car = Car()
    garage = Garage(mincepy.ObjRef(car))
    gid = garage.save()

    car_graph = historian.archive.get_obj_ref_graph(car.obj_id, direction=mincepy.INCOMING)
    assert len(car_graph.edges) == 1
    assert len(car_graph.nodes) == 2
    assert (gid, car.obj_id) in car_graph.edges

    # Now, create a new garage
    garage2 = Garage(mincepy.ObjRef(car))
    g2id = garage2.save()

    # Check that the reference graph is correct
    car_graph = historian.archive.get_obj_ref_graph(car.obj_id, direction=mincepy.INCOMING)
    assert len(car_graph.nodes) == 3
    assert len(car_graph.edges) == 2
    assert (gid, car.obj_id) in car_graph.edges
    assert (g2id, car.obj_id) in car_graph.edges

    # Finally, set the references to None
    garage.car = mincepy.ObjRef()
    garage.save()
    garage2.car = mincepy.ObjRef()
    garage2.save()

    # Check that the reference graph is correct
    car_graph = historian.archive.get_obj_ref_graph(gid)
    assert len(car_graph.edges) == 0
    assert len(car_graph.nodes) == 1


def test_obj_ref_max_depth(historian: mincepy.Historian):
    """Test object references max depth"""
    # Set up the chain: three -> two -> one -> zero
    zero = mincepy.ObjRef()
    one = mincepy.ObjRef(zero)
    two = mincepy.ObjRef(one)
    three = mincepy.ObjRef(two)
    zero_id, one_id, two_id, three_id = historian.save(zero, one, two, three)

    # No max depth
    graph = historian.archive.get_obj_ref_graph(two_id)
    assert len(graph.nodes) == 3
    assert len(graph.edges) == 2
    assert (one_id, zero_id) in graph.edges
    assert (two_id, one_id) in graph.edges

    graph = historian.archive.get_obj_ref_graph(three_id, max_dist=3)
    assert len(graph.nodes) == 4
    assert len(graph.edges) == 3
    assert (one_id, zero_id) in graph.edges
    assert (two_id, one_id) in graph.edges
    assert (three_id, two_id) in graph.edges

    graph = historian.archive.get_obj_ref_graph(two_id, max_dist=1)
    assert len(graph.edges) == 1
    assert (two_id, one_id) in graph.edges

    graph = historian.archive.get_obj_ref_graph(two_id, max_dist=0)
    assert len(graph.nodes) == 1
    assert len(graph.edges) == 0


def test_obj_referencing_max_depth(historian: mincepy.Historian):
    """Test object referencing max depth"""
    # Set up the chain: three -> two -> one -> zero
    zero = mincepy.ObjRef()
    one = mincepy.ObjRef(zero)
    two = mincepy.ObjRef(one)
    three = mincepy.ObjRef(two)
    zero_id, one_id, two_id, three_id = historian.save(zero, one, two, three)

    # No max depth
    graph = historian.archive.get_obj_ref_graph(one_id, direction=mincepy.INCOMING)
    assert len(graph.edges) == 2
    assert len(graph.nodes) == 3
    assert (three_id, two_id) in graph.edges
    assert (two_id, one_id) in graph.edges

    graph = historian.archive.get_obj_ref_graph(zero_id, direction=mincepy.INCOMING, max_dist=2)
    assert len(graph.edges) == 2
    assert len(graph.nodes) == 3
    assert (one_id, zero_id) in graph.edges
    assert (two_id, one_id) in graph.edges

    graph = historian.archive.get_obj_ref_graph(one_id, direction=mincepy.INCOMING, max_dist=1)
    assert len(graph.edges) == 1
    assert len(graph.nodes) == 2
    assert (two_id, one_id) in graph.edges

    graph = historian.archive.get_obj_ref_graph(two_id, direction=mincepy.INCOMING, max_dist=0)
    assert len(graph.edges) == 0
    assert len(graph.nodes) == 1



"""Module for testing saved snapshots"""

import time

import pytest

import mincepy
from mincepy.testing import Car, Cycle

# pylint: disable=invalid-name


def test_list_basics(historian: mincepy.Historian):
    parking_lot = mincepy.builtins.List()
    for i in range(1000):
        parking_lot.append(Car(str(i)))

    historian.save(parking_lot)
    list_sid = historian.get_snapshot_id(parking_lot)

    # Change one element
    parking_lot[0].make = 'ferrari'
    historian.save(parking_lot)
    new_list_sid = historian.get_snapshot_id(parking_lot)

    assert list_sid != new_list_sid

    old_list = historian.load_snapshot(list_sid)
    assert old_list is not parking_lot

    assert old_list[0].make == str(0)


def test_save_snapshot_change_load(historian: mincepy.Historian):
    car = Car()

    # Saving twice without changing should produce the same snapshot id
    historian.save(car)
    ferrari_sid = historian.get_snapshot_id(car)
    historian.save(car)
    assert ferrari_sid == historian.get_snapshot_id(car)

    car.make = 'fiat'
    car.color = 'white'

    historian.save(car)
    fiat_sid = historian.get_snapshot_id(car)

    assert fiat_sid != ferrari_sid

    ferrari = historian.load_snapshot(ferrari_sid)

    assert ferrari.make == 'ferrari'
    assert ferrari.colour == 'red'


def test_transaction_snapshots(historian: mincepy.Historian):
    ferrari = Car('ferrari')
    historian.save(ferrari)
    ferrari_sid = historian.get_snapshot_id(ferrari)

    with historian.transaction():
        ferrari_snapshot_1 = historian.load_snapshot(ferrari_sid)
        with historian.transaction():
            ferrari_snapshot_2 = historian.load_snapshot(ferrari_sid)
            # Reference wise they should be unequal
            assert ferrari_snapshot_1 is not ferrari_snapshot_2
            assert ferrari is not ferrari_snapshot_1
            assert ferrari is not ferrari_snapshot_2

            # Value wise they should be equal
            assert ferrari == ferrari_snapshot_1
            assert ferrari == ferrari_snapshot_2

        # Now check within the same transaction the result is the same
        ferrari_snapshot_2 = historian.load_snapshot(ferrari_sid)
        # Reference wise they should be unequal
        assert ferrari_snapshot_1 is not ferrari_snapshot_2
        assert ferrari is not ferrari_snapshot_1
        assert ferrari is not ferrari_snapshot_2

        # Value wise they should be equal
        assert ferrari == ferrari_snapshot_1
        assert ferrari == ferrari_snapshot_2


def test_record_times(historian: mincepy.Historian):
    car = Car('honda', 'red')
    historian.save(car)
    time.sleep(0.001)

    car.colour = 'white'
    historian.save(car)
    time.sleep(0.01)

    car.colour = 'yellow'
    historian.save(car)

    history = historian.history(car, as_objects=False)
    assert len(history) == 3
    for idx, record in zip(range(1, 2), history[1:]):
        assert record.creation_time == history[0].creation_time
        assert record.snapshot_time > history[0].creation_time
        assert record.snapshot_time > history[idx - 1].snapshot_time


def test_get_latest(historian: mincepy.Historian):
    # Save the car
    car = Car()
    car_id = historian.save(car)

    # Change it and save getting a snapshot
    car.make = 'fiat'
    car.colour = 'white'
    historian.save(car)
    fiat_sid = historian.get_snapshot_id(car)
    assert car_id != fiat_sid

    # Change it again...
    car.make = 'honda'
    car.colour = 'wine red'
    historian.save(car)
    honda_sid = historian.get_snapshot_id(car)
    assert honda_sid != fiat_sid
    assert honda_sid != car_id

    # Now delete and reload
    del car
    latest = historian.load(car_id)
    assert latest == historian.load_snapshot(honda_sid)


def test_history(historian: mincepy.Historian):
    rainbow = ['red', 'orange', 'yellow', 'green', 'blue', 'indigo', 'violet']

    car = Car()
    car_id = None
    for colour in rainbow:
        car.colour = colour
        car_id = historian.save(car)

    car_history = historian.history(car_id)
    assert len(car_history) == len(rainbow)
    for i, entry in enumerate(car_history):
        assert entry[1].colour == rainbow[i]

    # Test loading directly from snapshot id
    assert historian.load_snapshot(car_history[2].ref) == car_history[2].obj

    # Test slicing
    assert historian.history(car_id, -1)[0].obj.colour == rainbow[-1]

    # Try changing history
    old_version = car_history[2].obj
    old_version.colour = 'black'
    with pytest.raises(mincepy.ModificationError):
        historian.save(old_version)


def test_loading_snapshot(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    historian.save(honda)
    white_honda_sid = historian.get_snapshot_id(honda)
    honda.colour = 'red'
    historian.save(honda)
    del honda

    with historian.transaction():
        white_honda = historian.load_snapshot(white_honda_sid)
        assert white_honda.colour == 'white'
        # Make sure that if we load it again we get a different object instance
        assert white_honda is not historian.load_snapshot(white_honda_sid)


def test_loading_snapshot_cycle(historian: mincepy.Historian):
    a = Cycle()
    a.ref = a  # Close the cycle
    historian.save(a)
    a_sid = historian.get_snapshot_id(a)
    del a

    loaded = historian.load_snapshot(a_sid)
    assert loaded.ref is loaded



from collections import namedtuple
import contextlib

try:
    from contextlib import nullcontext
except ImportError:
    from contextlib2 import nullcontext
import getpass
import logging
import socket
import typing
from typing import MutableMapping, Any, Optional, Iterable, Union, Iterator, Type
import weakref

import deprecation

from . import archives
from . import builtins
from . import defaults
from . import depositors
from . import refs
from . import exceptions
from . import helpers
from . import hist
from . import migrate
from . import operations
from . import records
from . import staging
from . import tracking
from . import types
from . import type_registry
from . import utils
from . import version as version_mod
from .transactions import RollbackTransaction, Transaction, LiveObjects

__all__ = 'Historian', 'ObjectEntry'

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

ObjectEntry = namedtuple('ObjectEntry', 'ref obj')
HistorianType = Union[helpers.TypeHelper, typing.Type[types.SavableObject]]


class Historian:  # pylint: disable=too-many-public-methods, too-many-instance-attributes
    """The historian acts as a go-between between your python objects and the archive which is
    a persistent store of the records.  It will keep track of all live objects (i.e. those that
    have active references to them) that have been loaded and/or saved as well as enabling the
    user to lookup objects in the archive."""

    def __init__(self, archive: archives.Archive, equators=()):
        self._archive = archive
        self._equator = types.Equator(defaults.get_default_equators() + equators)
        # Register default types
        self._type_registry = type_registry.TypeRegistry()
        self.register_type(refs.ObjRef)
        self.register_type(builtins.SnapshotIdHelper())
        self.register_types(archive.get_types())

        # Snapshot objects -> reference. Objects that were loaded from historical snapshots
        self._snapshots_objects = utils.WeakObjectIdDict(
        )  # type: MutableMapping[Any, records.SnapshotId]
        self._live_objects = LiveObjects()

        self._transactions = None

        self._saving_set = set()

        self._user = getpass.getuser()
        self._hostname = socket.gethostname()

        self._live_depositor = depositors.LiveDepositor(self)
        self._meta = hist.Meta(self, self._archive)
        self._migrate = migrate.Migrations(self)
        self._references = hist.References(self)

    @property
    def archive(self):
        return self._archive

    @property
    def meta(self) -> hist.Meta:
        """Access to functions that operate on the metadata"""
        return self._meta

    @property
    def primitives(self) -> tuple:
        """A tuple of all the primitive types"""
        return types.PRIMITIVE_TYPES + (self._archive.get_id_type(),)

    @property
    def migrations(self) -> migrate.Migrations:
        """Access the migration possibilities"""
        return self._migrate

    @property
    def references(self) -> hist.References:
        """Access the references possibilities"""
        return self._references

    def create_file(self, filename: str = None, encoding: str = None) -> builtins.BaseFile:
        """Create a new file.  The historian will supply file type compatible with the archive in
         use."""
        return self._archive.create_file(filename, encoding)

    def save(self, *objs):
        """Save multiple objects producing corresponding object identifiers.  This returns a
        sequence of ids that is in the same order as the passed objects.

        :param objs: the object(s) to save.  Can also be a tuple of (obj, meta) to optionally
            include metadata to be saved with the object(s)
        """
        to_save = []
        # Convert everything to tuples
        for entry in objs:
            if isinstance(entry, tuple):
                if len(entry) > 2:
                    raise ValueError(
                        "Supplied tuples can only contain (object, meta), got '{}'".format(entry))
            else:
                entry = (entry,)
            to_save.append(entry)

        ids = []
        with self.in_transaction():
            for entry in to_save:
                ids.append(self.save_one(*entry))

        if len(objs) == 1:
            return ids[0]

        return ids

    def save_one(self, obj, meta: dict = None):
        """Save the object returning an object id.  If metadata is supplied it will be set on the
        object.

        Developer note: this is the front end point-of-entry for a user/client code saving an object
        however subsequent objects being saved in this transaction will only go through _save_object
        and therefore any code common to all objects being saved should possibly go there.
        """
        if obj in self._snapshots_objects:
            raise exceptions.ModificationError(
                "Cannot save a snapshot object, that would rewrite history!")

        if meta and not isinstance(meta, dict):
            raise TypeError("Metadata must be a dictionary, got type '{}'".format(type(meta)))

        # Save the object and metadata
        with self.in_transaction():
            record = self._save_object(obj, self._live_depositor)
            if meta:
                self.meta.update(record.obj_id, meta)

        return record.obj_id

    def is_known(self, obj) -> bool:
        """Check if an object has ever been saved and is therefore known to the historian

        :return: True if ever saved, False otherwise
        """
        return self.get_obj_id(obj) is not None

    def replace(self, old, new):
        """Replace a live object with a new version.

        This is especially useful if you have made a copy of an object and modified it but you want
        to continue the history of the object as the original rather than a brand new object.  Then
        just replace the old object with the new one by calling this function.
        """
        if self.current_transaction() is not None:
            raise RuntimeError("Can't replace during a transaction for the time being")

        if not isinstance(new, type(old)):
            raise TypeError("Can't replace type '{} with type '{}!".format(type(old), type(new)))

        # Get the current record and replace the object with the new one
        record = self._live_objects.get_record(old)
        self._live_objects.remove(record.obj_id)
        self._live_objects.insert(new, record)

        # Make sure creators is correct as well
        staging.replace(old, new)

    def load_snapshot(self, snapshot_id: records.SnapshotId) -> Any:
        return self._new_snapshot_depositor().load(snapshot_id)

    def load(self, *obj_ids_or_refs):
        """Load object(s) or snapshot(s)."""
        loaded = []
        for entry in obj_ids_or_refs:
            loaded.append(self.load_one(entry))

        if len(obj_ids_or_refs) == 1:
            return loaded[0]

        return loaded

    def load_one(self, obj_id_or_ref):
        """Load one object or shot from the database"""
        if isinstance(obj_id_or_ref, records.SnapshotId):
            return self.load_snapshot(obj_id_or_ref)

        return self._load_object(obj_id_or_ref, self._live_depositor)

    def sync(self, obj) -> bool:
        """Update an object with the latest state in the database.
        If there is no new version in the archive then the current version remains
        unchanged including any modifications.

        :return: True if the object was updated, False otherwise
        """
        obj_id = self.get_obj_id(obj)
        if obj_id is None:
            # Never saved so the object is as up to date as can be!
            return False

        sid = self._get_latest_snapshot_reference(obj_id)

        try:
            record = next(self._archive.find(obj_id=obj_id, version=-1))
        except StopIteration:
            raise exceptions.NotFound(obj_id)

        if record.is_deleted_record():
            raise exceptions.ObjectDeleted("Object with id '{}' has been deleted".format(obj_id))

        if sid.version == self.get_snapshot_id(obj).version:
            # Nothing has changed
            return False

        # The one in the archive is newer, so use that
        return self._live_depositor.update_from_record(obj, record)

    @deprecation.deprecated(deprecated_in="0.14.5",
                            removed_in="0.16",
                            current_version=version_mod.__version__,
                            details="Use mincepy.copy() instead")
    def copy(self, obj):  # pylint: disable=no-self-use
        """Create a shallow copy of the object.  Using this method allows the historian to inject
        information about where the object was copied from into the record if saved."""
        return tracking.copy(obj)

    def delete(self, obj_or_identifier):
        """Delete an object.

        :raises mincepy.NotFound: if the object cannot be found (potentially because it was
            already deleted)
        """
        # We need a record to be able to build the delete record
        obj_id = self._ensure_obj_id(obj_or_identifier)

        try:
            record = self.get_current_record(self.get_obj(obj_id))
        except exceptions.ObjectDeleted:
            # Object deleted already do reraise
            raise
        except exceptions.NotFound as exc:
            # Have a look in the archive
            results = self.archive.find(obj_id=obj_id, version=-1)
            try:
                record = next(results)
            except StopIteration:
                # not even in the archive
                raise exc

        with self.in_transaction() as trans:
            builder = records.make_deleted_builder(record)
            deleted_record = self._record_builder_created(builder).build()
            trans.delete(obj_id)
            trans.stage(operations.Insert(deleted_record))

    def history(
            self,
            obj_or_obj_id,
            idx_or_slice='*',
            as_objects=True) -> [typing.Sequence[ObjectEntry], typing.Sequence[records.DataRecord]]:
        """
        Get a sequence of object ids and instances from the history of the given object.

        :param obj_or_obj_id: The instance or id of the object to get the history for
        :param idx_or_slice: The particular index or a slice of which historical versions to get
        :param as_objects: if True return the object instances, otherwise returns the DataRecords

        Example:
        >>> historian = get_historian()
        >>> car = Car('ferrari', 'white')
        >>> car_id = historian.save(car)
        >>> car.colour = 'red'
        >>> historian.save(car)
        >>> history = historian.history(car_id)
        >>> len(history)
        2
        >>> history[0].obj.colour == 'white'
        True
        >>> history[1].obj.colour == 'red'
        True
        >>> history[1].obj is car
        """
        obj_id = self._ensure_obj_id(obj_or_obj_id)
        snapshot_refs = self._archive.get_snapshot_ids(obj_id)
        indices = utils.to_slice(idx_or_slice)
        to_get = snapshot_refs[indices]
        if as_objects:
            return [ObjectEntry(ref, self.load_snapshot(ref)) for ref in to_get]

        return [self._archive.load(ref) for ref in to_get]

    def get_current_record(self, obj) -> records.DataRecord:
        """Get a record for an object known to the historian"""
        trans = self.current_transaction()
        # Try the transaction first
        if trans:
            try:
                return trans.get_record_for_live_object(obj)
            except exceptions.ObjectDeleted:
                # ObjectDeleted is a specialisation of the NotFound error but it means that we
                # should consider the object as being gone so reraise it
                raise
            except exceptions.NotFound:
                pass

        return self._live_objects.get_record(obj)

    def get_obj_id(self, obj) -> Any:
        """Get the object ID for a live object.

        :return: the object id or None if the object is not known to the historian
        """
        trans = self.current_transaction()
        if trans is not None:
            try:
                return trans.get_snapshot_id_for_live_object(obj).obj_id
            except exceptions.NotFound:
                pass

        try:
            obj_id = self._live_objects.get_record(obj).obj_id
            if trans is not None and trans.is_deleted(obj_id):
                # The object has been deleted in the transaction, so it is not known
                return None

            return obj_id
        except exceptions.NotFound:
            return None

    def get_obj(self, obj_id):
        """Get a currently live object"""
        trans = self.current_transaction()
        if trans:
            try:
                return trans.get_live_object(obj_id)
            except exceptions.ObjectDeleted:
                # ObjectDeleted is a specialisation of the NotFound error but it means that we
                # should consider the object as being gone so reraise it
                raise
            except exceptions.NotFound:
                pass

        return self._live_objects.get_object(obj_id)

    def is_saved(self, obj) -> bool:
        """Test if an object is saved with this historian.  This is equivalent to
        `historian.get_obj_id(obj) is not None`."""
        return self.get_obj_id(obj) is not None

    def to_obj_id(self, obj_or_identifier):
        """
        This call will try and get an object id from the passed parameter.  There are three
        possibilities:

        1. Passed an object ID in which case it will be returned unchanged
        2. Passed a live object instance, in which case the id of that object will be returned
        3. Passed a type that can be understood by the archive as an object id e.g. a string of
           version, in which case the archive will attempt to convert it

        Returns None if neither of these cases were True.
        """
        if self.is_obj_id(obj_or_identifier):
            return obj_or_identifier

        try:
            # Try creating it for the user by calling the constructor with the argument passed.
            # This helps for common obj id types which can be constructed from a string
            return self._archive.construct_archive_id(obj_or_identifier)
        except (ValueError, TypeError):
            # Maybe we've been passed an object
            pass

        try:
            return self.get_obj_id(obj_or_identifier)
        except exceptions.NotFound:
            pass

        # Out of options
        return None

    def get_snapshot_id(self, obj) -> records.SnapshotId:
        """Get the current snapshot id for a live object.  Will return the id or raise
        :class:`mincepy.NotFound` exception"""
        trans = self.current_transaction()
        if trans:
            try:
                return trans.get_snapshot_id_for_live_object(obj)
            except exceptions.NotFound:
                pass

        return self._live_objects.get_record(obj).snapshot_id

    def hash(self, obj):
        return self._equator.hash(obj)

    def eq(self, one, other):  # pylint: disable=invalid-name
        return self._equator.eq(one, other)

    # region Types

    @classmethod
    def is_trackable(cls, obj):
        """Determine if an object is trackable i.e. we can treat these as live objects and
        automatically keep track of their history when saving.  Ultimately this is determined by
        whether the type is weak referencable or not.
        """
        try:
            weakref.ref(obj)
            return True
        except TypeError:
            return False

    @property
    def type_registry(self) -> type_registry.TypeRegistry:
        return self._type_registry

    def is_primitive(self, obj) -> bool:
        """Check if the object is one of the primitives and should be saved by value in the
        archive"""
        return isinstance(obj, self.primitives)

    def is_obj_id(self, obj_id) -> bool:
        """Check if an object is of the object id type"""
        return isinstance(obj_id, self._archive.get_id_type())

    def register_type(self, obj_class_or_helper: HistorianType) -> helpers.TypeHelper:
        helper = self._type_registry.register_type(obj_class_or_helper)
        self._equator.add_equator(helper)
        return helper

    def register_types(self, obj_clases_or_helpers: Iterable[HistorianType]):
        for item in obj_clases_or_helpers:
            self.register_type(item)

    def get_obj_type_id(self, obj_type):
        return self._type_registry.get_type_id(obj_type)

    def get_obj_type(self, type_id):
        return self.get_helper(type_id).TYPE

    def get_helper(self, type_id_or_type, auto_register=False) -> helpers.TypeHelper:
        if auto_register and issubclass(type_id_or_type, types.SavableObject):
            self._ensure_compatible(type_id_or_type)

        return self._type_registry.get_helper(type_id_or_type)

    # endregion

    def find(self,
             obj_type=None,
             obj_id=None,
             version: int = -1,
             state=None,
             meta: dict = None,
             sort=None,
             limit=0,
             skip=0) -> Iterator[Any]:
        """
        .. _MongoDB: https://docs.mongodb.com/manual/tutorial/query-documents/

        Find objects.  This call will search the archive for objects matching the given criteria.
        In many cases the main arguments of interest will be `state` and `meta` which allow you to
        apply filters on the stored state of the object and metadata respectively.  To understand
        how the state is stored in the database (and therefore how to apply filters to it) it may
        be necessary to look at the details of the `save_instance_state()` method for that type.
        Metadata is always a dictionary containing primitives (strings, dicts, lists, etc).

        For the most part, the filter syntax of `mincepy` conforms to that of `MongoDB`_ with
        convenience functions locate in :py:mod:`mincepy.qops` that can make it easier to
        to build a query.

        Examples:

        Find all :py:class:`~mincepy.testing.Car`s that are brown or red:

        >>> import mincepy
        >>> historian.find(state=dict(colour=mincepy.q.in_('brown', 'red')))

        Find all people that are older than 34 and live in Edinburgh:

        >>> historian.find(state=dict(age=mincepy.q.gt_(34)), meta=dict(city='Edinburgh'))

        :param obj_type: the object type to look for
        :param obj_id: an object or multiple object ids to look for
        :param version: the version of the object to retrieve, -1 means latest
        :param state: the criteria on the state of the object to apply
        :type state: must be subclass of historian.primitive
        :param meta: the search criteria to apply on the metadata of the object
        :param sort: the sort criteria
        :param limit: the maximum number of results to return, 0 means unlimited
        :param skip: the page to get results from
        """
        # pylint: disable=too-many-arguments
        results = self.find_records(obj_type=obj_type,
                                    obj_id=obj_id,
                                    version=version,
                                    state=state,
                                    meta=meta,
                                    sort=sort,
                                    limit=limit,
                                    skip=skip)
        for result in results:
            yield self.load(result.obj_id)

    def find_records(self,
                     obj_type=None,
                     obj_id=None,
                     version: int = -1,
                     state=None,
                     meta: dict = None,
                     extras: dict = None,
                     sort=None,
                     limit=0,
                     skip=0) -> Iterator[records.DataRecord]:
        """Find records

        :param obj_type: the object type to look for
        :param obj_id: an object or multiple object ids to look for
        :param version: the version of the object to retrieve, -1 means latest
        :param state: the criteria on the state of the object to apply
        :type state: must be subclass of historian.primitive
        :param meta: the search criteria to apply on the metadata of the object
        :param extras: the search criteria to apply on the data record extras
        :param sort: the sort criteria
        :param limit: the maximum number of results to return, 0 means unlimited
        :param skip: the page to get results from
        """
        # pylint: disable=too-many-arguments
        type_id = obj_type
        if obj_type is not None:
            try:
                type_id = self.get_obj_type_id(obj_type)
            except TypeError:
                pass
        results = self._archive.find(obj_id=obj_id,
                                     type_id=type_id,
                                     state=state,
                                     version=version,
                                     meta=meta,
                                     extras=extras,
                                     sort=sort,
                                     limit=limit,
                                     skip=skip)
        yield from results

    def get_creator(self, obj_or_identifier):
        """Get the object that created the passed object"""
        if not self.is_obj_id(obj_or_identifier):
            # Object instance, try the staging area
            info = staging.get_info(obj_or_identifier, create=False) or {}
            created_by = info.get(records.ExtraKeys.CREATED_BY, None)
            if created_by is not None:
                return created_by

        creator_id = self.created_by(obj_or_identifier)
        return self.load_one(creator_id)

    def created_by(self, obj_or_identifier):
        """Return the id of the object that created the passed object"""
        try:
            record = self.get_current_record(obj_or_identifier)
        except exceptions.NotFound:
            if not self.is_obj_id(obj_or_identifier):
                raise

            try:
                record = next(self._archive.find(obj_id=obj_or_identifier, version=-1))
            except StopIteration:
                raise exceptions.NotFound(obj_or_identifier)

        return record.created_by

    def get_user_info(self) -> dict:
        """Get information about the current user and host"""
        user_info = {}
        if self._user:
            user_info[records.ExtraKeys.USER] = self._user
        if self._hostname:
            user_info[records.ExtraKeys.HOSTNAME] = self._hostname
        return user_info

    @contextlib.contextmanager
    def in_transaction(self):
        """This context will either re-use an existing transaction, if one is currently taking place
        or create a new one if not."""
        current = self.current_transaction()
        if current is None:
            ctx = self.transaction()
        else:
            ctx = nullcontext(current)

        with ctx as trans:
            yield trans

    @contextlib.contextmanager
    def transaction(self):
        """Start a new transaction.  Will be nested if there is already one underway"""
        if self._transactions:
            # Start a nested one
            with self._transactions[-1].nested() as nested:
                self._transactions.append(nested)
                try:
                    yield nested
                except Exception:  # Need this so we can have 'else' pylint: disable=try-except-raise
                    raise
                else:
                    self._closing_transaction(nested)
                finally:
                    popped = self._transactions.pop()
                    assert popped is nested
        else:
            # New transaction
            trans = Transaction()
            self._transactions = [trans]

            try:
                yield trans
            except RollbackTransaction:
                pass
            else:
                self._closing_transaction(trans)
                self._commit_transaction(trans)
            finally:
                assert len(self._transactions) == 1
                assert self._transactions[0] is trans
                self._transactions = None

    def current_transaction(self) -> Optional[Transaction]:
        """Get the current transaction if there is one, otherwise returns None"""
        if not self._transactions:
            return None
        return self._transactions[-1]

    def _closing_transaction(self, trans: Transaction):
        conflicting = set()

        # Filter out the deleted records
        del_ops = filter(
            lambda _: isinstance(_, operations.Insert) and _.record.is_deleted_record(),
            trans.staged)

        obj_ids = set(operation.obj_id for operation in del_ops)
        ref_graph = self.references.get_obj_ref_graph(*obj_ids, direction=archives.INCOMING)
        for obj_id in obj_ids:
            for edge in ref_graph.in_edges(obj_id):
                conflicting.add(edge[1])

        if conflicting:
            raise exceptions.ReferenceError("Cannot perform delete", conflicting)

    def _commit_transaction(self, trans: Transaction):
        """Commit a transaction that is finishing"""
        # Live objects
        self._live_objects.update(trans.live_objects)
        # Deleted objects
        for deleted in trans.deleted:
            try:
                self._live_objects.remove(deleted)
            except exceptions.NotFound:
                pass

        # Snapshots
        for ref, obj in trans.snapshots.items():
            self._snapshots_objects[obj] = ref

        # Save any records that were staged for archiving
        if trans.staged:
            self._archive.bulk_write(trans.staged)

        # Metas
        if trans.metas:
            self._archive.meta_set_many(trans.metas)

    def _get_latest_snapshot_reference(self, obj_id) -> records.SnapshotId:
        """Given an object id this will return a reference to the latest snapshot"""
        try:
            return self._archive.get_snapshot_ids(obj_id)[-1]
        except IndexError:
            raise exceptions.NotFound("Object with id '{}' not found.".format(obj_id))

    def _load_object(self, obj_id, depositor: depositors.LiveDepositor):
        obj_id = self._ensure_obj_id(obj_id)

        with self.in_transaction() as trans:
            # Try getting the object from the our dict of up to date ones
            try:
                return trans.get_live_object(obj_id)
            except exceptions.NotFound:
                pass

            if trans.is_deleted(obj_id):
                raise exceptions.ObjectDeleted(obj_id)

            # Couldn't find it, so let's check if we have one and check if it is up to date
            results = tuple(self.archive.find(obj_id, version=-1))
            if not results:
                raise exceptions.NotFound(obj_id)
            record = results[0]

            if record.is_deleted_record():
                raise exceptions.ObjectDeleted(obj_id)

            try:
                obj = self._live_objects.get_object(obj_id)
            except exceptions.NotFound:
                logger.debug("Loading object from record: %s", record.snapshot_id)
                # Ok, just use the one from the archive
                return depositor.load_from_record(record)
            else:
                if record.version != self.get_snapshot_id(obj).version:
                    # The one in the archive is newer, so use that
                    logger.debug("Updating object from record: %s", record.snapshot_id)
                    depositor.update_from_record(obj, record)

                return obj

    def _save_object(self, obj, depositor) -> records.DataRecord:
        with self.in_transaction() as trans:
            try:
                helper = self._ensure_compatible(type(obj))
            except TypeError:
                raise TypeError("Object is incompatible with the historian: {}".format(obj))

            # Check if an object is already being saved in the transaction
            try:
                record = trans.get_record_for_live_object(obj)
                return record
            except exceptions.NotFound:
                pass

            with self._saving(obj):
                # Ok, have to save it
                current_hash = self.hash(obj)

                try:
                    # Let's see if we have a record at all
                    record = self._live_objects.get_record(obj)
                except exceptions.NotFound:
                    # Object being saved for the first time
                    builder = self._create_builder(obj, snapshot_hash=current_hash)
                    record = depositor.save_from_builder(obj, builder)
                    if self.meta.sticky:
                        # Apply the sticky meta
                        self.meta.update(record.obj_id, self.meta.sticky)
                    return record
                else:
                    if helper.IMMUTABLE:
                        logger.info("Tried to save immutable object with id '%s' again",
                                    record.obj_id)
                        return record

                    # Check if our record is up to date
                    with self.transaction() as nested:
                        loaded_obj = self._new_snapshot_depositor().load_from_record(record)

                        if current_hash == record.snapshot_hash and self.eq(obj, loaded_obj):
                            # Objects identical
                            nested.rollback()
                        else:
                            builder = record.child_builder(snapshot_hash=current_hash)
                            record = depositor.save_from_builder(obj, builder)

                    return record

    def _create_builder(self, obj, **additional) -> records.DataRecordBuilder:
        """Create a record builder for a new object object"""
        additional = additional or {}
        helper = self._ensure_compatible(type(obj))

        builder = records.DataRecord.new_builder(type_id=helper.TYPE_ID,
                                                 obj_id=self._archive.create_archive_id(),
                                                 version=0)
        builder.update(additional)
        return builder

    def _ensure_obj_id(self, obj_or_identifier):
        """
        This call will try and get an object id from the passed parameter.  There are three
        possibilities:
            1. Passed an object ID in which case it will be returned unchanged
            2. Passed a live object instance, in which case the id of that object will be returned
            3. Passed a type that can be understood by the archive as an object id e.g. a string of
               version, in which case the archive will attempt to convert it
        """
        obj_id = self.to_obj_id(obj_or_identifier)
        if obj_id is None:
            raise exceptions.NotFound(
                "Could not get an object id from '{}'".format(obj_or_identifier))

        return obj_id

    def _ensure_compatible(self, obj_type: Type[types.SavableObject]) -> helpers.TypeHelper:
        if obj_type not in self._type_registry:
            self.register_type(obj_type)

        return self._type_registry.get_helper_from_obj_type(obj_type)

    @contextlib.contextmanager
    def _saving(self, obj):
        obj_id = id(obj)
        if obj_id in self._saving_set:
            raise RuntimeError(
                "The object is already being saved, this cannot be called twice and suggests "
                "a circular reference is being made")
        self._saving_set.add(obj_id)
        try:
            yield
        finally:
            self._saving_set.remove(obj_id)

    def _record_builder_created(self,
                                builder: records.DataRecordBuilder) -> records.DataRecordBuilder:
        """Update a data record builder with standard information."""
        builder.extras.update(self.get_user_info())
        return builder

    def _new_snapshot_depositor(self):
        return depositors.SnapshotLoader(self)



""""Tests of migration"""
import gc

import mincepy
from mincepy import testing
from .common import CarV1, CarV2, StoreByValue, StoreByRef


def test_find_migratable(historian: mincepy.Historian):
    car = CarV1('white', 'lada')
    car_id = car.save()
    by_val = StoreByValue(car)
    by_val_id = by_val.save()

    # Register a new version of the car
    historian.register_type(CarV2)

    # Now both car and by_val should need migration (because by_val stores a car)
    migratable = tuple(historian.migrations.find_migratable_records())
    assert len(migratable) == 2
    ids = [record.obj_id for record in migratable]
    assert car_id in ids
    assert by_val_id in ids

    # Now register a new version of StoreByVal
    historian.register_type(StoreByRef)

    # There should still be the same to migratables as before
    migratable = tuple(historian.migrations.find_migratable_records())
    assert len(migratable) == 2
    ids = [record.obj_id for record in migratable]
    assert car_id in ids
    assert by_val_id in ids


def test_migrate_with_saved(historian: mincepy.Historian):
    """Test migrating an object that has saved references"""

    class V3(mincepy.SimpleSavable):
        ATTRS = (mincepy.AsRef('ref'), 'description')
        TYPE_ID = StoreByRef.TYPE_ID

        class Migration(mincepy.ObjectMigration):
            VERSION = 2
            PREVIOUS = StoreByRef.ToRefMigration

            @classmethod
            def upgrade(cls, saved_state, loader: 'mincepy.Loader'):
                saved_state['description'] = None
                return saved_state

        LATEST_MIGRATION = Migration

        def __init__(self, ref):
            super().__init__()
            self.ref = ref
            self.description = None

    obj = StoreByRef(testing.Car())
    obj_id = obj.save()
    del obj
    gc.collect()

    historian.register_type(V3)
    migrated = historian.migrations.migrate_all()
    assert len(migrated) == 1
    assert migrated[0].obj_id == obj_id

    obj = historian.load(obj_id)

    assert isinstance(obj, V3)
    assert hasattr(obj, 'description')
    assert obj.description is None



__all__ = ('NotFound', 'ModificationError', 'ObjectDeleted', 'DuplicateKeyError', 'MigrationError',
           'VersionError', 'IntegrityError', 'ReferenceError', 'ConnectionError')


class ConnectionError(Exception):  # pylint: disable=redefined-builtin
    """Raise when there is an error in connecting to the backend"""


class NotFound(Exception):
    """Raised when something can not be found in the history"""


class ModificationError(Exception):
    """Raised when a modification of the history encountered a problem"""


class ObjectDeleted(NotFound):
    """Raise when the user tries to interact with a deleted object"""


class DuplicateKeyError(Exception):
    """Indicates that a uniqueness constraint was violated"""


class MigrationError(Exception):
    """Indicates that an error occurred during migration"""


class VersionError(Exception):
    """Indicates a version mismatch between the code and the database"""


class IntegrityError(Exception):
    """Indicates an error that occurred because of an operation that would conflict with a database
    constraint"""


class ReferenceError(IntegrityError):  # pylint: disable=redefined-builtin

    def __init__(self, msg, references: set):
        super().__init__(msg)
        self.references = references



from typing import Optional

import pymongo.database

from . import db


def get_settings(database: pymongo.database.Database, name: str = db.GLOBAL_SETTINGS) -> \
        Optional[dict]:
    return database[db.SETTINGS_COLLECTION].find_one(name)


def set_settings(database: pymongo.database.Database,
                 settings: Optional[dict],
                 name: str = db.GLOBAL_SETTINGS):
    coll = database[db.SETTINGS_COLLECTION]

    if settings is None:
        coll.delete_one({'_id': name})
    else:
        coll.replace_one({'_id': name}, settings, upsert=True)



"""Module containing record operations that can be performed sent to the archive to perform"""
import abc

from . import records

__all__ = 'Operation', 'Insert', 'Update', 'Delete'


class Operation(metaclass=abc.ABCMeta):
    """Base class for all operations"""

    @property
    @abc.abstractmethod
    def obj_id(self):
        """The is of the object being operated on"""

    @property
    @abc.abstractmethod
    def snapshot_id(self):
        """The snapshot id of the object being operated on"""


class Insert(Operation):
    """Insert a record into the archive

    For use with :meth:`~mincepy.Archive.bulk_write`
    """

    def __init__(self, record: records.DataRecord):
        self._record = record

    @property
    def obj_id(self):
        return self._record.obj_id

    @property
    def snapshot_id(self):
        return self._record.snapshot_id

    @property
    def record(self) -> records.DataRecord:
        return self._record


class Update(Operation):
    """Update a record currently in the archive.  This takes the snapshot id and a dictionary
    containing the fields to be updated.  The update operation behaves like a dict.update()"""

    def __init__(self, sid: records.SnapshotId, update: dict):
        diff = set(update.keys()) - set(records.DataRecord._fields)
        if diff:
            raise ValueError("Invalid keys found in the update operation: {}".format(diff))

        self._sid = sid
        self._update = update

    @property
    def obj_id(self):
        return self._sid.obj_id

    @property
    def snapshot_id(self) -> records.SnapshotId:
        """The snapshot being updated"""
        return self._sid

    @property
    def update(self) -> dict:
        """The update that will be performed"""
        return self._update


class Delete(Operation):
    """Delete a record from the archive"""

    def __init__(self, sid: records.SnapshotId):
        self._sid = sid

    @property
    def obj_id(self):
        return self._sid.obj_id

    @property
    def snapshot_id(self) -> records.SnapshotId:
        """The snapshot being deleted"""
        return self._sid



import collections
from typing import Type, MutableMapping, Iterable, Any, Union

from . import helpers
from . import types

SavableObjectType = Type[types.SavableObject]


class TypeRegistry:
    """The type registry.  This contains helpers that furnish mincepy with the necessary methods
    to store and track objects in the archive"""

    def __init__(self):
        self._helpers = {}  # type: MutableMapping[SavableObjectType, helpers.TypeHelper]
        self._type_ids = {}  # type: MutableMapping[Any, SavableObjectType]

    def __contains__(self, item: SavableObjectType) -> bool:
        return item in self._helpers

    @property
    def type_helpers(self) -> MutableMapping[Type, helpers.TypeHelper]:
        """Get the mapping of registered type helpers"""
        return self._helpers

    def register_type(
            self, obj_class_or_helper: Union[helpers.TypeHelper, SavableObjectType]) \
            -> helpers.WrapperHelper:
        """Register a type new type"""
        helper = self._register(obj_class_or_helper)

        # Now, put in any ancestors
        for ancestor in reversed(types.savable_mro(helper.TYPE)[1:]):
            self._register(ancestor)

        return helper

    def get_type_id(self, obj_type: SavableObjectType):
        try:
            # Try a direct lookup first
            return self._helpers[obj_type].TYPE_ID
        except KeyError:
            # Try an issubclass lookup as a backup
            for type_id, known_type in self._type_ids.items():
                if issubclass(obj_type, known_type):
                    return type_id
            raise ValueError("Type '{}' is not known".format(obj_type))

    def get_helper(self, type_id_or_type) -> helpers.TypeHelper:
        if isinstance(type_id_or_type, type):
            return self.get_helper_from_obj_type(type_id_or_type)

        return self.get_helper_from_type_id(type_id_or_type)

    def get_helper_from_type_id(self, type_id) -> helpers.TypeHelper:
        try:
            return self.get_helper_from_obj_type(self._type_ids[type_id])
        except KeyError:
            raise TypeError("Type id '{}' not known".format(type_id))

    def get_helper_from_obj_type(self, obj_type: SavableObjectType) -> helpers.TypeHelper:
        try:
            # Try the direct lookup
            return self._helpers[obj_type]
        except KeyError:
            # Do the full issubclass lookup
            for known_type, helper in self._helpers.items():
                if issubclass(obj_type, known_type):
                    return helper
            raise ValueError("Type '{}' has not been registered".format(obj_type))

    def get_version_info(self, type_id_or_type) -> collections.OrderedDict:
        """Get version information about a type.  This will return a reverse mro ordered dictionary
        where the key is the type id and the value is the version.  Only registered entries will
        appear."""
        helper = self.get_helper(type_id_or_type)
        type_info = collections.OrderedDict()

        # Get information for myself
        type_info[helper.TYPE_ID] = helper.get_version()

        # Get information for everything in the mro
        for type_id in reversed(helper.TYPE.mro()):
            try:
                helper = self.get_helper(type_id)
            except ValueError:
                pass
            else:
                type_info[helper.TYPE_ID] = helper.get_version()

        return type_info

    def _register(
            self, obj_class_or_helper: Union[helpers.TypeHelper, SavableObjectType]) \
            -> helpers.WrapperHelper:
        """Register a type and return the associated helper"""
        if isinstance(obj_class_or_helper, helpers.TypeHelper):
            helper = obj_class_or_helper
        else:
            if not issubclass(obj_class_or_helper, types.Object):
                raise TypeError("Type '{}' is nether a TypeHelper nor a SavableObject".format(
                    obj_class_or_helper))
            helper = helpers.WrapperHelper(obj_class_or_helper)

        self._insert_helper(helper)
        return helper

    def _insert_helper(self, helper: helpers.TypeHelper):
        """Insert a helper into the registry for all the types that it supports"""
        obj_types = helper.TYPE if isinstance(helper.TYPE, Iterable) else (helper.TYPE,)

        for obj_type in obj_types:
            self._helpers[obj_type] = helper
            self._type_ids[helper.TYPE_ID] = obj_type



"""Classes and function useful for trying out mincepy functionality"""
import os
import random
import uuid

import bson
import pymongo
import pytest

import mincepy

ENV_ARCHIVE_URI = 'MINCEPY_TEST_URI'
DEFAULT_ARCHIVE_URI = 'mongodb://localhost/mincepy-tests'

# pylint: disable=redefined-outer-name


@pytest.fixture
def archive_uri():
    return os.environ.get(ENV_ARCHIVE_URI, DEFAULT_ARCHIVE_URI)


@pytest.fixture
def mongodb_archive(archive_uri):
    client = pymongo.MongoClient(archive_uri)
    db = client.get_default_database()  # pylint: disable=invalid-name
    client.drop_database(db)
    mongo_archive = mincepy.mongo.MongoArchive(db)
    yield mongo_archive
    client.drop_database(db)


@pytest.fixture(autouse=True)
def historian(mongodb_archive):
    hist = mincepy.Historian(mongodb_archive)
    hist.register_types(mincepy.testing.HISTORIAN_TYPES)
    mincepy.set_historian(hist)
    yield hist
    mincepy.set_historian(None)


class Car(mincepy.SimpleSavable):
    TYPE_ID = bson.ObjectId('5e075d6244572f823ed93274')
    ATTRS = 'colour', 'make'

    def __init__(self, make='ferrari', colour='red'):
        super(Car, self).__init__()
        self.make = make
        self.colour = colour


class Garage(mincepy.SimpleSavable):
    TYPE_ID = bson.ObjectId('5e07b40a44572f823ed9327b')
    ATTRS = ('car',)

    def __init__(self, car=None):
        super(Garage, self).__init__()
        self.car = car

    def save_instance_state(self, saver: mincepy.Saver):
        state = super().save_instance_state(saver)
        return state


class Person(mincepy.SimpleSavable):
    TYPE_ID = uuid.UUID('d60ca740-9fa6-4002-83f6-e4c91403e41b')
    ATTRS = 'name', 'age', mincepy.AsRef('car')

    def __init__(self, name, age, car=None):
        super(Person, self).__init__()
        self.name = name
        self.age = age
        self.car = car


class Cycle(mincepy.SimpleSavable):
    TYPE_ID = uuid.UUID('600fb6ae-684c-4f8e-bed3-47ae06739d29')
    ATTRS = ('_ref',)

    def __init__(self, ref=None):
        super(Cycle, self).__init__()
        self._ref = mincepy.ObjRef(ref)

    @property
    def ref(self):
        return self._ref()

    @ref.setter
    def ref(self, value):
        self._ref = mincepy.ObjRef(value)

    def __eq__(self, other):
        return self.ref is other.ref

    def yield_hashables(self, hasher):
        yield from hasher.yield_hashables(id(self.ref))


def populate(historian=None):
    historian = historian or mincepy.get_historian()

    colours = ('red', 'orange', 'yellow', 'green', 'blue', 'indigo', 'violet')
    makes = ('honda', 'ferrari', 'zonda', 'fiat')

    cars = []

    for make in makes:
        for colour in colours:
            # Make some cars
            car = Car(make, colour)
            historian.save(car)
            cars.append(car)

    # Now randomly change some of them
    for _ in range(int(len(cars) / 4)):
        car = random.choice(cars)
        car.colour = random.choice(colours)
        car.save()

    # Now change one a number of times
    car = random.choice(cars)
    for colour in colours:
        car.colour = colour
        car.save()

    people = mincepy.RefList()
    for name in ('martin', 'sonia', 'gavin', 'upul', 'sebastiaan', 'irene'):
        person = Person(name, random.randint(20, 40))
        historian.save(person)
        people.append(person)
    historian.save(people)


HISTORIAN_TYPES = Car, Garage, Person, Cycle



"""References module"""
from typing import Optional

from . import exceptions
from . import types
from . import records

__all__ = ('ObjRef',)


class ObjRef(types.SavableObject):
    """A reference to an object instance"""
    TYPE_ID = records.SnapshotId.TYPE_ID
    IMMUTABLE = True

    _obj = None
    _sid = None  # type: Optional[records.SnapshotId]
    _loader = None

    def __init__(self, obj=None):
        super().__init__()
        assert not (obj is not None and types.is_primitive(obj)), \
            "Can't create a reference to a primitive type"
        self._obj = obj

    def __bool__(self) -> bool:
        """Test if this is a null reference"""
        return self._obj is not None or self._sid is not None

    def __str__(self) -> str:
        desc = ["ObjRef('"]
        if self._obj is not None:
            desc.append(str(self._obj))
        else:
            desc.append(str(self._sid))
        desc.append("')")
        return "".join(desc)

    def __repr__(self) -> str:
        return "ObjRef({})".format(self._obj if self._obj is not None else self._sid)

    def __call__(self, update=False):
        """Get the object being referenced.  If update is called then the latest version
        will be loaded from the historian"""
        if self._obj is None:
            # This means we were loaded and need to load the object
            if self._sid is None:
                raise RuntimeError("Cannot dereference a None reference")
            # Cache the object
            self._obj = self._loader.load(self._sid)
            assert self._obj is not None, "Loader did not load object"
            self._sid = None
            self._loader = None
        elif update:
            try:
                self._historian.sync(self._obj)
            except exceptions.NotFound:
                pass  # Object must never have been saved and is therefore up to date

        return self._obj

    def __eq__(self, other) -> bool:
        if not isinstance(other, ObjRef):
            return False

        if self._obj is not None:
            return id(self._obj) == id(other._obj)

        return self._sid == other._sid

    def yield_hashables(self, hasher):
        if self._obj is not None:
            yield from hasher.yield_hashables(id(self._obj))
        else:
            # This will also work if ref is None
            yield from hasher.yield_hashables(self._sid)

    def save_instance_state(self, saver):
        if self._obj is not None:
            sid = saver.get_snapshot_id(self._obj)
        else:
            sid = self._sid

        if sid is not None:
            return sid.to_dict()

        return None

    def load_instance_state(self, saved_state, loader):
        super(ObjRef, self).load_instance_state(saved_state, loader)
        # Rely on class default values for members
        if saved_state is not None:
            if isinstance(saved_state, list):
                # Legacy version
                self._sid = records.SnapshotId(*saved_state)
            else:
                # New version is dict
                self._sid = records.SnapshotId(**saved_state)
            self._loader = loader


HISTORIAN_TYPES = (ObjRef,)



from typing import MutableMapping, Any, Optional

from . import utils

__all__ = 'get_info', 'remove'


class StagingArea:
    """This global singleton stores information about objects before they have been saved by the
    historian.  This is useful when there is global information that may or may not be used if the
    object is indeed saved.  If not, the information is simply discarded when the object is
    destructed"""

    # pylint: disable=too-few-public-methods

    staged_obj_info = utils.WeakObjectIdDict()  # type: MutableMapping[Any, dict]

    def __init__(self):
        raise RuntimeError("Cannot be instantiated")


def get_info(obj, create=True) -> Optional[dict]:
    """Get the information dictionary for a given staged object.  If create is True, the
    dictionary will be created if it doesn't already exist.  If False and the object is not
    staged then None will be returned.
    """
    if create:
        return StagingArea.staged_obj_info.setdefault(obj, {})

    try:
        return StagingArea.staged_obj_info[obj]
    except KeyError:
        return None


def remove(obj):
    """Remove the information dictionary for a staged object.  If the object is not staged
    then this function does nothing"""
    StagingArea.staged_obj_info.pop(obj)


def replace(old, new):
    """Move over the staging information from object 'old' to object 'new'"""
    try:
        staging_info = StagingArea.staged_obj_info.pop(old)
    except KeyError:
        # Delete any information present for new
        StagingArea.staged_obj_info.pop(new, None)
    else:
        StagingArea.staged_obj_info[new] = staging_info



import logging
from typing import List

import stevedore

logger = logging.getLogger(__name__)


def get_types() -> List:
    """Get all mincepy types and type helper instances registered as extensions"""
    mgr = stevedore.extension.ExtensionManager(
        namespace='mincepy.plugins.types',
        invoke_on_load=False,
    )

    all_types = []

    def get_type(extension: stevedore.extension.Extension):
        try:
            all_types.extend(extension.plugin())
        except Exception:
            logger.exception("Failed to get types plugin from %s", extension.name)

    mgr.map(get_type)

    return all_types



import mincepy
from mincepy.testing import Car


def test_meta_set_update_many(historian: mincepy.Historian):
    car1 = Car()
    car2 = Car()
    car1id, car2id = historian.save(car1, car2)
    historian.archive.meta_set_many({car1id: {'reg': 'car1'}, car2id: {'reg': 'car2'}})

    results = historian.archive.meta_get_many((car1id, car2id))
    assert results == {car1id: {'reg': 'car1'}, car2id: {'reg': 'car2'}}

    historian.archive.meta_update_many({car1id: {'colour': 'red'}, car2id: {'reg': 'car2updated'}})

    metas = historian.archive.meta_get_many((car1id, car2id))
    assert metas == {car1id: {'reg': 'car1', 'colour': 'red'}, car2id: {'reg': 'car2updated'}}


def test_meta_find(historian: mincepy.Historian):
    car1 = Car()
    car2 = Car()

    car1id, _ = historian.save(car1, car2)
    historian.archive.meta_set(car1id, {'reg': 'car1'})

    results = dict(historian.archive.meta_find({}, (car1id,)))
    assert results == {car1id: {'reg': 'car1'}}


def test_meta_update_many(historian: mincepy.Historian):
    car1 = Car()
    car2 = Car()
    car1id, car2id = historian.save(car1, car2)
    historian.archive.meta_set_many({car1id: {'reg': 'car1'}, car2id: {'reg': 'car2'}})

    results = historian.archive.meta_get_many((car1id, car2id))
    assert results[car1id] == {'reg': 'car1'}
    assert results[car2id] == {'reg': 'car2'}


def test_count(historian: mincepy.Historian):
    car1 = Car('ferrari')
    car2 = Car('skoda')
    car1id, car2id = historian.save(car1, car2)
    historian.archive.meta_set_many({car1id: {'reg': 'car1'}, car2id: {'reg': 'car2'}})

    assert historian.archive.count() == 2
    assert historian.archive.count(state=dict(make='ferrari')) == 1
    assert historian.archive.count(meta=dict(reg='car1')) == 1


def test_find_from_id(historian: mincepy.Historian):
    car = Car('ferrari')
    car_id = car.save()

    results = tuple(historian.archive.find(obj_id=car_id))
    assert len(results) == 1
    assert results[0].obj_id == car_id

    # Now check that we can pass an iterable of ids
    car2 = Car('skoda')
    car2_id = car2.save()
    results = tuple(historian.archive.find(obj_id=[car_id, car2_id]))
    assert len(results) == 2
    ids = [record.obj_id for record in results]
    assert car_id in ids
    assert car2_id in ids


def test_find_using_iterator(mongodb_archive: mincepy.Archive):
    """Test that passing an iterable to find types that support it, works."""
    record_details = dict(state=None, state_types=None, snapshot_hash=None)

    record1 = mincepy.DataRecord.new_builder(obj_id=123, type_id=1, **record_details).build()
    record2 = mincepy.DataRecord.new_builder(obj_id=456, type_id=2, **record_details).build()

    mongodb_archive.save_many([record1, record2])

    results = tuple(mongodb_archive.find(obj_id=[123].__iter__()))
    assert len(results) == 1

    results = tuple(mongodb_archive.find(type_id=[1].__iter__()))
    assert len(results) == 1



import logging
import sys

import click
import pytray.pretty
import tabulate

import mincepy.plugins

__all__ = ('mince',)

# pylint: disable=import-outside-toplevel


def set_print_logging(level=logging.WARNING):
    mince_logger = logging.getLogger('mincepy')
    mince_logger.setLevel(level)

    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    mince_logger.addHandler(handler)


@click.group()
def mince():
    pass


@mince.command()
@click.argument('uri', type=str)
def gui(uri):
    """Start the mincepy gui (mincepy_gui package must be installed)"""
    try:
        import mincepy_gui
    except ImportError:
        click.echo("mincepy-gui not found, please install (e.g. via pip install mincepy-gui)")
        sys.exit(1)
    else:
        mincepy_gui.start(uri)


@mince.command()
def plugins():
    type_plugins = mincepy.plugins.get_types()
    click.echo("Types:")

    headers = 'Type', 'Class'
    plugin_info = []
    for plugin in type_plugins:
        if isinstance(plugin, mincepy.TypeHelper):
            row = 'Type Helper', pytray.pretty.type_string(type(plugin))
        elif isinstance(plugin, type) and issubclass(plugin, mincepy.SavableObject):
            row = 'Savable Object', pytray.pretty.type_string(plugin)
        else:
            row = 'Unrecognised!', pytray.pretty.type_string(plugin)
        plugin_info.append(row)

    click.echo(tabulate.tabulate(plugin_info, headers))


@mince.command()
@click.option('--yes', is_flag=True, default=False, help="Yes to all prompts")
@click.argument('uri', type=str)
@click.pass_context
def migrate(ctx, yes, uri):
    try:
        hist = mincepy.create_historian(uri)
    except ValueError as exc:
        click.echo(exc)
        sys.exit(1)
    else:
        if isinstance(ctx.obj, dict) and 'helpers' in ctx.obj:
            hist.register_types(ctx.obj['helpers'])

        click.echo("Looking for records to migrate...")
        records = tuple(hist.migrations.find_migratable_records())
        if not records:
            click.echo("No migrations necessary")
            return

        click.echo("Found {} records to migrate".format(len(records)))
        if yes or click.confirm("Migrate all?"):
            set_print_logging(logging.INFO)
            hist.migrations.migrate_records(records)
        else:
            sys.exit(2)



import uuid

import pytest

import mincepy
from mincepy.testing import Car

# pylint: disable=invalid-name


def test_metadata_simple(historian: mincepy.Historian):
    car = Car()
    ferrari_id = historian.save((car, {'reg': 'VD395'}))
    # Check that we get back what we just set
    assert historian.meta.get(ferrari_id) == {'reg': 'VD395'}

    car.make = 'fiat'
    red_fiat_id = historian.save(car)
    # Check that the metadata is shared
    assert historian.meta.get(red_fiat_id) == {'reg': 'VD395'}

    historian.meta.set(ferrari_id, {'reg': 'N317'})
    # Check that this saves the metadata on the object level i.e. both are changed
    assert historian.meta.get(ferrari_id) == {'reg': 'N317'}
    assert historian.meta.get(red_fiat_id) == {'reg': 'N317'}


def test_metadata_using_object_instance(historian: mincepy.Historian):
    car = Car()
    historian.save((car, {'reg': 'VD395'}))
    # Check that we get back what we just set
    assert historian.meta.get(car) == {'reg': 'VD395'}

    car.make = 'fiat'
    historian.save(car)
    # Check that the metadata is shared
    assert historian.meta.get(car) == {'reg': 'VD395'}

    historian.meta.set(car, {'reg': 'N317'})
    assert historian.meta.get(car) == {'reg': 'N317'}


def test_metadata_multiple(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    zonda = Car('zonda', 'yellow')

    historian.save((honda, {'reg': 'H123'}), (zonda, {'reg': 'Z456'}))

    assert historian.meta.get(honda) == {'reg': 'H123'}
    assert historian.meta.get(zonda) == {'reg': 'Z456'}


def test_metadata_wrong_type(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    with pytest.raises(TypeError):
        historian.save_one(honda, meta=['a', 'b'])


def test_metadata_update(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    historian.save_one(honda, meta=dict(reg='H123', vin=1234, owner='Mike'))
    historian.meta.update(honda, {'vin': 5678, 'owner': 'Mart'})
    assert historian.meta.get(honda) == {'reg': 'H123', 'vin': 5678, 'owner': 'Mart'}


def test_metadata_update_inexistant(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    historian.save(honda)
    # If the data doesn't exist in the metadata already we expect an update to simply insert
    historian.meta.update(honda, {'reg': 'H123', 'vin': 1234})
    assert historian.meta.get(honda) == {'reg': 'H123', 'vin': 1234}


def test_metadata_find_objects(historian: mincepy.Historian):
    honda = Car('honda', 'white')
    honda2 = Car('honda', 'white')
    historian.save_one(honda, meta={'reg': 'H123', 'vin': 1234})
    historian.save(honda2)

    results = list(historian.find(Car, meta={'reg': 'H123'}))
    assert len(results) == 1


# region sticky-meta


def test_stick_meta(historian: mincepy.Historian):
    car1 = Car()
    historian.meta.sticky['owner'] = 'martin'
    car2 = Car()
    car3 = Car()

    car1_id = car1.save()
    car2_id = car2.save(meta={'for_sale': True})
    car3_id = car3.save(meta={'owner': 'james'})
    del car1, car2, car3

    assert historian.meta.get(car1_id) == {'owner': 'martin'}
    assert historian.meta.get(car2_id) == {'owner': 'martin', 'for_sale': True}
    assert historian.meta.get(car3_id) == {'owner': 'james'}


def test_meta_sticky_children(historian: mincepy.Historian):
    """Catch bug where metadata was not being set on being references by other objects"""
    garage = mincepy.RefList()
    garage.append(Car())
    garage.append(Car())
    historian.meta.sticky['owner'] = 'martin'

    garage_id = garage.save()
    car0_id = garage[0].save(meta={'for_sale': True})
    car1_id = garage[1].save(meta={'owner': 'james'})
    del garage

    assert historian.meta.get(garage_id) == {'owner': 'martin'}
    assert historian.meta.get(car0_id) == {'owner': 'martin', 'for_sale': True}
    assert historian.meta.get(car1_id) == {'owner': 'james'}


def test_meta_stick_copy(historian: mincepy.Historian):
    """Test that sticky meta is applied to objects that are copied.  Issue #13:
    https://github.com/muhrin/mincepy/issues/13"""
    car = Car()
    historian.meta.sticky.update({'owner': 'martin'})
    car.save()
    car_copy = mincepy.copy(car)
    car_copy.save()

    assert historian.meta.get(car) == {'owner': 'martin'}
    assert historian.meta.get(car_copy) == {'owner': 'martin'}


# endregion


def test_meta_transaction(historian: mincepy.Historian):
    """Check that metadata respects transaction boundaries"""
    car1 = Car()

    with historian.transaction():
        car1.save()
        with historian.transaction() as trans:
            historian.meta.set(car1.obj_id, {'spurious': True})
            assert historian.meta.get(car1.obj_id) == {'spurious': True}
            trans.rollback()
        assert not historian.meta.get(car1)
    assert not historian.meta.get(car1)


def test_metadata_find_object_regex(historian: mincepy.Historian):
    car1 = Car('honda', 'white')
    car2 = Car('honda', 'white')
    car3 = Car('honda', 'white')

    car1.save(meta={'reg': 'VD395'})
    car2.save(meta={'reg': 'VD574'})
    car3.save(meta={'reg': 'BE368'})

    # Find all cars with a reg starting in VD
    results = tuple(historian.find(Car, meta={'reg': {'$regex': '^VD'}}))
    assert len(results) == 2
    assert results[0] in [car1, car2]
    assert results[1] in [car1, car2]


def test_metadata_find(historian: mincepy.Historian):
    """Test querying for metadata directly"""
    car1 = Car('honda', 'white')
    car2 = Car('honda', 'white')
    car3 = Car('honda', 'white')

    car1.save(meta={'reg': 'VD395'})
    car2.save(meta={'reg': 'VD574'})
    car3.save(meta={'reg': 'BE368'})

    results = dict(historian.meta.find(filter=dict(reg={'$regex': '^VD'})))
    assert len(results) == 2
    ids = results.keys()
    assert car1.obj_id in ids
    assert car2.obj_id in ids


def test_meta_delete(historian: mincepy.Historian):
    """Test that when an object is delete so is its metadata"""
    car = Car()
    car_id = car.save(meta={'reg': '1234'})
    results = dict(historian.meta.get(car))
    assert results == {'reg': '1234'}

    historian.delete(car)

    assert historian.meta.get(car_id) is None

    results = tuple(historian.meta.find({}, obj_id=car_id))
    assert len(results) == 0


def test_set_meta_in_save(historian: mincepy.Historian):

    class Info(mincepy.BaseSavableObject):
        TYPE_ID = uuid.UUID('6744689d-5f88-482e-bb42-2bec5f139cc2')

        def save_instance_state(self, saver: mincepy.Saver) -> dict:
            state = super().save_instance_state(saver)
            saver.get_historian().meta.set(self, dict(msg='good news'))
            return state

    info = Info()
    historian.save(info)

    meta = historian.meta.get(info)
    assert meta == {'msg': 'good news'}


def test_update_meta_in_save(historian: mincepy.Historian):

    class Info(mincepy.SimpleSavable):
        TYPE_ID = uuid.UUID('6744689d-5f88-482e-bb42-2bec5f139cc2')

        def save_instance_state(self, saver: mincepy.Saver) -> dict:
            state = super().save_instance_state(saver)
            saver.get_historian().meta.update(self, dict(msg='good news'))
            return state

    info = Info()
    info.save()

    meta = historian.meta.get(info)
    assert meta == {'msg': 'good news'}


def test_update_meta_in_save_with_sticky(historian: mincepy.Historian):

    class Info(mincepy.SimpleSavable):
        TYPE_ID = uuid.UUID('6744689d-5f88-482e-bb42-2bec5f139cc2')

        def save_instance_state(self, saver: mincepy.Saver) -> dict:
            state = super().save_instance_state(saver)
            saver.get_historian().meta.update(self, dict(msg='good news'))
            return state

    historian = mincepy.get_historian()
    historian.meta.sticky['all'] = 'good'

    info = Info()
    info.save()

    meta = historian.meta.get(info)
    assert meta == {'all': 'good', 'msg': 'good news'}


def test_set_meta_in_save_fail(historian: mincepy.Historian):
    """Make sure that metadata isn't saved if saving of the object fails"""

    class Info(mincepy.SimpleSavable):
        TYPE_ID = uuid.UUID('6744689d-5f88-482e-bb42-2bec5f139cc2')

        def save_instance_state(self, saver: mincepy.Saver) -> dict:
            super().save_instance_state(saver)
            self.set_meta(dict(msg='good news'))
            raise RuntimeError("I'm crashin' yo")

    info = Info()
    with pytest.raises(RuntimeError):
        info.save()

    with pytest.raises(mincepy.NotFound):
        historian.meta.get(info)


def test_meta_unique_index(historian: mincepy.Historian):
    historian.meta.create_index('reg', True)
    car = Car()
    car.save(meta={'reg': 'VD495'})

    historian.archive.meta_find(dict(reg='VD495'))

    with pytest.raises(mincepy.DuplicateKeyError):
        Car().save(meta={'reg': 'VD495'})


def test_meta_index_unique_joint_index_where_exist(historian: mincepy.Historian):
    """Test a joint index that allows missing entries"""
    historian.meta.create_index([('reg', mincepy.ASCENDING), ('colour', mincepy.ASCENDING)],
                                unique=True,
                                where_exist=True)

    Car().save(meta=dict(reg='VD395', colour='red'))

    Car().save(meta=dict(colour='red'))
    Car().save(meta=dict(colour='red'))  # This should be ok

    Car().save(meta=dict(reg='VD395'))
    Car().save(meta=dict(reg='VD395'))  # This should be ok

    with pytest.raises(mincepy.DuplicateKeyError):
        Car().save(meta=dict(reg='VD395', colour='red'))


def test_meta_on_delete(historian: mincepy.Historian):
    """Test that metadata gets deleted when the object does"""
    car = Car()
    car_id = car.save(meta={'reg': '1234'})

    assert historian.meta.get(car_id) == {'reg': '1234'}
    historian.delete(car_id)
    assert historian.meta.get(car_id) is None


def test_meta_distinct(historian: mincepy.Historian):
    car1 = Car()
    car2 = Car()
    car3 = Car()

    historian.save(car1, car2, car3)

    car1.set_meta({'owner': 'martin'})
    car2.set_meta({'owner': 'bob'})
    car3.set_meta({'owner': 'martin'})

    assert set(historian.meta.distinct('owner')) == {'martin', 'bob'}



# -*- coding: utf-8 -*-
from setuptools import setup

__author__ = "Martin Uhrin"
__license__ = "GPLv3 and MIT"

about = {}
with open('mincepy/version.py') as f:
    exec(f.read(), about)

setup(
    name='mincepy',
    version=about['__version__'],
    description="Python object storage with versioning made simple",
    long_description=open('README.rst').read(),
    url='https://github.com/muhrin/mincepy.git',
    author='Martin Uhrin',
    author_email='martin.uhrin.10@ucl.ac.uk',
    license=__license__,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='database schemaless nosql orm object-store concurrent optimistic-locking',
    install_requires=[
        'contextlib2; python_version<"3.7"',
        'deprecation',
        'pymongo',
        'bidict',
        'networkx',  # For reference graphs
        'pyblake2; python_version<"3.6"',
        'pytray>=0.2.1',
        'stevedore',
        'click',
        'tabulate',
    ],
    extras_require={
        'cli': ['click', 'tabulate'],
        'gui': ['mincepy-gui'],
        'dev': [
            'ipython',
            'pip',
            'pytest>4',
            'pytest-benchmark',
            'pytest-cov',
            'pre-commit',
            'prospector',
            'pylint',
            'twine',
            'yapf',
        ],
        'docs': [
            'nbsphinx',
            'sphinx',
            'sphinx-autobuild',
        ],
    },
    packages=['mincepy', 'mincepy.cli', 'mincepy.mongo', 'mincepy.hist'],
    include_package_data=True,
    test_suite='test',
    provides=['mincepy.plugins'],
    entry_points={
        'console_scripts': ['mince = mincepy.cli.main:mince'],
        'mincepy.plugins.types': ['native_types = mincepy.provides:get_types'],
    })



import mincepy.records
from mincepy.testing import *


def test_find_state(historian: mincepy.Historian):
    honda = Car('honda', 'green')
    porsche = Car('porsche', 'black')
    red_honda = Car('honda', 'red')
    fiat = Car('fiat', 'green')

    historian.save(honda, porsche, red_honda, fiat)
    hondas = list(historian.find(Car, state={'make': 'honda'}))
    assert len(hondas) == 2
    assert honda in hondas
    assert red_honda in hondas

    # Try without type
    greens = list(historian.find(state={'colour': 'green'}))
    assert len(greens) == 2
    assert honda in greens
    assert fiat in greens


def test_find_skip(historian: mincepy.Historian):
    cars = []
    for idx in range(10):
        cars.append(Car(idx))

    historian.save(*cars)
    # Try live
    makes = set(range(10))
    # Skip two at a time limiting to two at a time, effectively paginating
    for skip in range(5):
        results = list(historian.find(obj_type=Car, limit=2, skip=skip * 2))
        assert len(results) == 2, "Got no results on page {}".format(skip)
        makes.remove(results[0].make)
        makes.remove(results[1].make)

    assert not makes


def test_find_by_type_id(historian: mincepy.Historian):
    """Test that searching by the type id works too"""
    cars = []
    for idx in range(10):
        cars.append(Car(idx))

    historian.save(*cars)
    result = tuple(historian.find(obj_type=Car.TYPE_ID))
    assert len(result) == 10


def test_simple_sort(historian: mincepy.Historian):
    cars = mincepy.builtins.RefList()
    for idx in range(10):
        cars.append(Car(idx))

    historian.save(cars)
    results = list(historian.find_records(Car, sort=mincepy.records.CREATION_TIME))
    for idx, result in enumerate(results[1:]):
        # No need to subtract 1 from idx as we're already one behind because of the slicing
        assert result.creation_time >= results[idx].creation_time


def test_find_latest(historian: mincepy.Historian):
    """Check that search criteria is only applied to the latest version if -1 is used"""
    # Version 0
    car = Car('fiat', 'yellow')
    car.save()
    # Version 1
    car.colour = 'white'
    car.save()
    # Version 2
    car.colour = 'red'
    car.save()

    car2 = Car('ford', 'blue')
    car2.save()
    car2.colour = 'green'
    car2.save()

    # Find the original version
    results = tuple(historian.archive.find(version=0, state=dict(colour='yellow')))
    assert len(results) == 1

    # Now search for a white fiat as the latest version
    results = tuple(historian.archive.find(version=-1, state=dict(colour='white')))
    assert len(results) == 0

    # Now do a version-unrestricted search
    results = tuple(historian.archive.find(version=None, state=dict(colour='yellow')))
    assert len(results) == 1
    assert results[0].version == 0

    # Now do a version-unrestricted search
    results = tuple(historian.archive.find(version=None, state=dict(make='fiat')))
    assert len(results) == 3


def test_find_predicates(historian: mincepy.Historian):
    """Test using the query predicates"""

    skoda = Car('skoda', 'green')
    ferrari = Car('ferrari', 'yellow')
    bugatti = Car('bugatti', 'green')
    historian.save(skoda, ferrari, bugatti)

    # OR
    results = tuple(
        historian.find_records(state=mincepy.q.or_({'make': 'skoda'}, {'make': 'ferrari'})))
    assert len(results) == 2
    makes = [record.state['make'] for record in results]
    assert 'skoda' in makes
    assert 'ferrari' in makes

    # AND
    results = tuple(
        historian.find_records(state=mincepy.q.and_({'make': 'skoda'}, {'colour': 'green'})))
    assert len(results) == 1
    makes = [record.state['make'] for record in results]
    assert 'skoda' in makes


@pytest.mark.skip(
    reason="This test takes a long time and consumes a lot of memory but if a MongoDB issue occurs "
    "that involves running out of memory it may be useful again")
def test_sort_with_many_entries(historian: mincepy.Historian):
    """Test for mincepy_gui issue #11 which was caused by a query with a large number of objects
    that was sorted.  This exceeded MongoDBs memory limits.  The fix is to simply allow disk use."""
    cars = []
    for i in range(1000000):
        cars.append(Car())

    historian.save(*cars)

    next(historian.find_records(Car, sort='state.make'))



import numbers
import collections.abc
from operator import itemgetter
import uuid

from . import helpers

__all__ = ('SimpleHelper', 'BytesEquator')


class SimpleHelper(helpers.TypeHelper):

    def yield_hashables(self, obj, hasher):
        yield from obj.yield_hashables(hasher)

    def eq(self, one, other) -> bool:
        return one == other

    def save_instance_state(self, obj, saver):
        return obj.save_instance_state(saver)

    def load_instance_state(self, obj, saved_state, loader):
        return obj.load(saved_state, loader)


class BytesEquator(SimpleHelper):
    TYPE = collections.abc.ByteString

    def yield_hashables(self, obj, hasher):
        yield obj


class StrEquator(SimpleHelper):
    TYPE = str

    def yield_hashables(self, obj: str, hasher):
        yield obj.encode('utf-8')


class SequenceEquator(SimpleHelper):
    TYPE = collections.abc.Sequence

    def yield_hashables(self, obj: collections.abc.Sequence, hasher):
        for entry in obj:
            yield from hasher.yield_hashables(entry)


class SetEquator(SimpleHelper):
    TYPE = collections.abc.Set

    def yield_hashables(self, obj: collections.abc.Set, hasher):
        for entry in sorted(obj):
            yield from hasher.yield_hashables(entry)


class MappingEquator(SimpleHelper):
    TYPE = collections.abc.Mapping

    def yield_hashables(self, obj, hasher):

        def hashed_key_mapping(mapping):
            for key, value in mapping.items():
                yield tuple(hasher.yield_hashables(key)), value

        for key_hashables, value in sorted(hashed_key_mapping(obj), key=itemgetter(0)):
            # Yield all the key hashables
            yield from key_hashables
            # And now all the value hashables for that entry
            yield from hasher.yield_hashables(value)


class OrderedDictEquator(SimpleHelper):
    TYPE = collections.OrderedDict

    def yield_hashables(self, obj, hasher):
        for key, val in sorted(obj, key=itemgetter(0)):
            yield from hasher.yield_hashables(key)
            yield from hasher.yield_hashables(val)


class RealEquator(SimpleHelper):
    TYPE = numbers.Real

    def yield_hashables(self, obj, hasher):
        yield from hasher.yield_hashables(hasher.float_to_str(obj))


class ComplexEquator(SimpleHelper):
    TYPE = numbers.Complex

    def yield_hashables(self, obj: numbers.Complex, hasher):
        yield from hasher.yield_hashables(obj.real)
        yield from hasher.yield_hashables(obj.imag)


class IntegerEquator(SimpleHelper):
    TYPE = numbers.Integral

    def yield_hashables(self, obj: numbers.Integral, hasher):
        yield from hasher.yield_hashables(u'{}'.format(obj))


class BoolEquator(SimpleHelper):
    TYPE = bool

    def yield_hashables(self, obj, hasher):
        yield b'\x01' if obj else b'\x00'


class NoneEquator(SimpleHelper):
    TYPE = type(None)

    def yield_hashables(self, obj, hasher):
        yield from hasher.yield_hashables('None')


class TupleEquator(SimpleHelper):
    TYPE = tuple

    def yield_hashables(self, obj, hasher):
        yield from hasher.yield_hashables(obj)


class UuidEquator(SimpleHelper):
    TYPE = uuid.UUID

    def yield_hashables(self, obj: uuid.UUID, hasher):
        yield obj.bytes



import itertools
import logging
from typing import Sequence, Union, Callable, Iterator, Iterable

import bson
import networkx
import pymongo.collection

import mincepy
from mincepy import OUTGOING
from mincepy import q
from . import db
from . import types

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class ReferenceManager:
    """Reference manager is able to find references between objects efficiently.

    This relies on the following assumptions:
        * The data collection _id field os the obj_id, and,
        * The history collection _id field is the snapshot id string
    """

    def __init__(self, ref_collection: pymongo.collection.Collection,
                 data_collection: pymongo.collection.Collection,
                 history_collection: pymongo.collection.Collection):
        self._references = ref_collection
        self._data_collection = data_collection
        self._history_collection = history_collection

    def get_obj_ref_graphs(
            self, obj_ids: Sequence[bson.ObjectId], direction=OUTGOING, max_dist: int = None) \
            -> networkx.DiGraph:
        return self._get_graph(obj_ids, direction=direction, max_dist=max_dist)

    def get_snapshot_ref_graph(
            self, ids: Sequence[mincepy.SnapshotId], direction=OUTGOING, max_dist: int = None) \
            -> Iterator[networkx.DiGraph]:
        """Get the reference graph for a sequence of ids"""
        return self._get_graph(ids,
                               direction=direction,
                               max_dist=max_dist,
                               node_factory=db.sid_from_str)

    def invalidate(self, obj_ids: Iterable[bson.ObjectId],
                   snapshot_ids: Iterable[types.SnapshotId]):
        """Invalidate the cache for these objects and snapshots"""
        to_delete = list(obj_ids)
        for sid in snapshot_ids:
            to_delete.append(str(sid))
        self._references.delete_many({'_id': q.in_(*to_delete)})

    def _get_graph(self,
                   ids: Sequence[Union[bson.ObjectId, types.SnapshotId]],
                   direction=OUTGOING,
                   max_dist: int = None,
                   node_factory: Callable = None) -> networkx.DiGraph:
        """Get the reference graph for a sequence of ids"""
        if max_dist == 0:
            # Special case for 0 distance, can't be any references just the ids as lone nodes
            graph = networkx.DiGraph()
            for entry_id in ids:
                graph.add_node(entry_id)
            return graph

        node_factory = node_factory or (lambda x: x)

        search_max_dist = max_dist
        if max_dist is not None and direction == OUTGOING:
            search_max_dist = max(max_dist - 1, 0)

        search_ids = self._prepare_for_ref_search(ids)
        pipeline = self._get_ref_pipeline(search_ids, direction=direction, max_dist=search_max_dist)
        ref_results = {result['_id']: result for result in self._references.aggregate(pipeline)}

        graph = networkx.DiGraph()

        for entry_id in search_ids:
            this_id = node_factory(entry_id)

            graph.add_node(this_id)
            result = ref_results.get(entry_id, None)
            if result:
                refs = result.get('references', [])

                # First add all the nodes
                for ref in refs:
                    neighbour = node_factory(ref['_id'])
                    graph.add_node(neighbour)

                # Then the edges
                for entry in itertools.chain([result], refs):
                    this = node_factory(entry['_id'])

                    for neighbour_id in entry['refs']:
                        neighbour = node_factory(neighbour_id)

                        if direction == OUTGOING or neighbour in graph.nodes:
                            graph.add_edge(this, neighbour)

        return graph

    def _get_ref_pipeline(self, ids: Sequence, direction=OUTGOING, max_dist: int = None) -> list:
        """Get the reference lookup pipeline."""
        if max_dist is not None and max_dist < 0:
            raise ValueError("max_dist must be positive, got '{}'".format(max_dist))

        pipeline = [{'$match': {'_id': q.in_(*ids)}}]

        if max_dist is None or max_dist != 0:
            lookup_params = {
                'from': self._references.name,
                'as': 'references',
                'depthField': 'depth'
            }
            if direction == OUTGOING:
                lookup_params.update(
                    dict(startWith='$refs', connectFromField='refs', connectToField='_id'))
            else:
                lookup_params.update(
                    dict(startWith='$_id', connectFromField='_id', connectToField='refs'))

            if max_dist is not None:
                lookup_params['maxDepth'] = max_dist - 1

            # Only do graph lookup if checking depth that involves a hop
            pipeline.append({'$graphLookup': lookup_params})

        return pipeline

    def _prepare_for_ref_search(self, ids: Sequence[Union[bson.ObjectId, mincepy.SnapshotId]]):
        hist_updated = False
        data_updated = False
        converted = []
        for entry in ids:
            if isinstance(entry, mincepy.SnapshotId):
                if not hist_updated:
                    self._ensure_current('history')
                    hist_updated = True

                # We use the string version of a snapshot id
                converted.append(str(entry))
            elif isinstance(entry, bson.ObjectId):
                if not data_updated:
                    self._ensure_current('data')
                    data_updated = True

                converted.append(entry)
            else:
                raise TypeError(entry)

        return converted

    def _ensure_current(self, collection_name: str):
        """This call ensures that the reference collection is up to date"""
        # Find all the objects that we don't have in the references collection
        if collection_name == 'history':
            collection = self._history_collection
            id_func = lambda schema_entry: str(schema_entry[1])
        elif collection_name == 'data':
            collection = self._data_collection
            id_func = lambda schema_entry: schema_entry[1].obj_id
        else:
            raise ValueError("Unsupported collection: {}".format(collection_name))

        to_insert = []
        for data_entry in self._get_missing_entries(collection):
            ref_entry = _generate_ref_entry(data_entry, id_func)
            ref_entry['_id'] = data_entry['_id']
            to_insert.append(ref_entry)

        # Now insert all the calculated refs
        if to_insert:
            logging.info("Updating references cache for collection '%s' with %i new entries",
                         collection_name, len(to_insert))
            self._references.insert_many(to_insert)

    def _get_missing_entries(self, collection: pymongo.collection.Collection) -> Iterator:
        """Given a collection get the records that do not have entries in the references cache"""
        pipeline = [{
            '$lookup': {
                'from': self._references.name,
                'localField': '_id',
                'foreignField': '_id',
                'as': 'references'
            }
        }, {
            '$match': {
                'references': []
            }
        }]
        yield from collection.aggregate(pipeline)


def _generate_ref_entry(data_entry: dict, id_func: Callable) -> dict:
    refs = [id_func(info) for info in db.to_record(data_entry).get_references()]
    return {'refs': refs}



import mincepy
from mincepy.testing import Car


def test_load(historian: mincepy.Historian):
    car = Car()
    car_id = historian.save(car)

    assert mincepy.load(car_id) is car

    # Check loading from 'cold'
    del car
    car = mincepy.load(car_id)
    assert car._historian == historian


def test_save(historian: mincepy.Historian):
    car = Car()
    mincepy.save(car)
    assert car._historian is historian



"""Module for testing saved snapshots"""

import mincepy
from mincepy import testing


def test_snapshot_id_in_transaction(historian: mincepy.Historian):
    """An object saved during a transaction should already have its new snapshot id available during
    the transaction."""
    car = testing.Car('ferrari', 'red')
    car.save()
    sid = historian.get_snapshot_id(car)

    with historian.transaction():
        car.make = 'honda'
        car.save()
        assert historian.get_snapshot_id(car) != sid
